import unittest
from fractions import Fraction

import numpy as np
import torch
from parameterized import parameterized

import neurosym as ns
from tests.utils import assertDSL

from .utils import ChildrenInOrderAsserterMask, ChildrenInOrderMask, ProbabilityTester


def get_dsl(with_vars=False, with_3=False):
    dslf = ns.DSLFactory()
    dslf.concrete("+", "(i, i) -> i", lambda x, y: x + y)
    dslf.concrete("1", "() -> i", lambda: 1)
    dslf.concrete("2", "() -> i", lambda: 2)
    if with_3:
        dslf.concrete("3", "() -> i", lambda: 3)
    if with_vars:
        dslf.concrete("call", "(i -> i, i) -> i", lambda f, x: f(x))
        dslf.lambdas()
    dslf.prune_to("i")
    return dslf.finalize()


def get_dsl_for_ordering():
    dslf = ns.DSLFactory()
    dslf.concrete("+", "(i, i, i) -> t", lambda x, y: x + y)
    dslf.concrete("1", "() -> i", lambda: 1)
    dslf.concrete("2", "() -> i", lambda: 2)
    dslf.concrete("3", "() -> i", lambda: 2)
    dslf.prune_to("t")
    return dslf.finalize()


dsl = get_dsl()
fam = ns.BigramProgramDistributionFamily(dsl)
dsl_with_3 = get_dsl(with_3=True)
fam_with_3 = ns.BigramProgramDistributionFamily(dsl_with_3)
fam_with_3_no_type_mask = ns.BigramProgramDistributionFamily(
    dsl_with_3, include_type_preorder_mask=False
)
dsl_with_vars = get_dsl(with_vars=True)
fam_with_vars = ns.BigramProgramDistributionFamily(dsl_with_vars)
dsl_for_ordering = get_dsl_for_ordering()
fam_with_ordering = ns.BigramProgramDistributionFamily(
    dsl_for_ordering, additional_preorder_masks=[ChildrenInOrderMask]
)
fam_with_ordering_asserted = ns.BigramProgramDistributionFamily(
    dsl_for_ordering, additional_preorder_masks=[ChildrenInOrderAsserterMask]
)
fam_with_ordering_231 = ns.BigramProgramDistributionFamily(
    dsl_for_ordering,
    additional_preorder_masks=[ChildrenInOrderMask],
    # 2 3 1 means go to index 2 first, then 0 second, then 1 third
    node_ordering=lambda dist: ns.DictionaryNodeOrdering(dist, {"+": [2, 0, 1]}),
)

uniform = [
    [
        # root -> each
        [0, 1 / 3, 1 / 3, 1 / 3],
        [0] * 4,
    ],
    [
        # + -> each first arg
        [0, 1 / 3, 1 / 3, 1 / 3],
        # + -> each second arg
        [0, 1 / 3, 1 / 3, 1 / 3],
    ],
    [[0] * 4] * 2,
    [[0] * 4] * 2,
]


class BigramSamplerTest(ProbabilityTester):
    def test_uniform_sample_counts(self):
        n = 10_000
        pdelta = 0.015
        dist = fam.uniform()
        self.assertSameProbDist(dist, uniform)
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertBinomial(n, 1 / 3, pdelta, samples.count("(1)"))
        for plus_two_things in (
            "(+ (1) (1))",
            "(+ (1) (2))",
            "(+ (2) (1))",
            "(+ (2) (2))",
        ):
            self.assertBinomial(
                n, 1 / 3 * 1 / 3 * 1 / 3, pdelta, samples.count(plus_two_things)
            )


class BigramParameterShapeTest(unittest.TestCase):
    def test_shape(self):
        self.assertEqual(fam.parameters_shape(), (4, 2, 4))


class BigramWithParametersGetParametersTest(ProbabilityTester):
    def test_produces_uniform_with_fixed_input(self):
        parameters = torch.zeros((4, 2, 4))
        for off in -1, 298, 29, 3:
            self.assertSameProbDist(
                fam.with_parameters((parameters + off)[None])[0], uniform
            )

    def test_irrelevant_entries_dont_matter(self):
        parameters = torch.zeros((4, 2, 4))
        parameters[0, 0, 0] = 1038
        print(parameters)
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

        parameters = torch.zeros((4, 2, 4))
        parameters[0, 1, 1] = -3
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

        parameters = torch.zeros((4, 2, 4))
        parameters[2] = 3
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

    def test_can_make_deterministic(self):
        parameters = torch.zeros((4, 2, 4))
        # + -> always 1 then 2
        parameters[1, 0, 2] = 100
        parameters[1, 1, 3] = 100

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)", "(+ (1) (2))"})

        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(2)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(+ (1) (2))"))

    def test_can_make_impossible(self):
        parameters = torch.zeros((4, 2, 4))
        # disable root -> +
        parameters[0, 0, 1] = -100

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)"})

        self.assertBinomial(n, 1 / 2, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 2, 0.015, samples.count("(2)"))

    def test_can_weight_precisely(self):
        parameters = torch.zeros((4, 2, 4))
        # disable root -> +
        parameters[0, 0, 1] = -100
        # make root -> 1 twice as likely as root -> 2
        parameters[0, 0, 2] = np.log(2)

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)"})

        self.assertBinomial(n, 2 / 3, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(2)"))

    def test_fam_with_variables_dsl(self):
        assertDSL(
            self,
            dsl_with_vars.render(),
            """
            $0_0 :: V<i@0>
            $1_0 :: V<i@1>
            $2_0 :: V<i@2>
            $3_0 :: V<i@3>
            + :: (i, i) -> i
            1 :: () -> i
            2 :: () -> i
            call :: (i -> i, i) -> i
            lam :: L<#body|i> -> i -> #body
            """,
        )

    def test_sample_with_variables(self):
        dist = fam_with_vars.uniform()
        n = 20000
        samples = [
            fam_with_vars.sample(dist, np.random.RandomState(i)) for i in range(n)
        ]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertBinomial(n, 1 / 4, 0.01, samples.count("(1)"))
        self.assertBinomial(n, 1 / 4, 0.01, samples.count("(2)"))
        # see test_call_with_variables for the math
        self.assertBinomial(n, 1 / 80, 0.01, samples.count("(call (lam ($0_0)) (1))"))

    def test_sample_with_ordering(self):
        dist = fam_with_ordering.uniform()
        n = 10_000
        samples = [
            fam_with_ordering.sample(dist, np.random.RandomState(i)) for i in range(n)
        ]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(+ (1) (2) (3))"})

    def test_sample_with_ordering_231(self):
        dist = fam_with_ordering_231.uniform()
        n = 10_000
        samples = [
            fam_with_ordering_231.sample(dist, np.random.RandomState(i))
            for i in range(n)
        ]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(+ (2) (3) (1))"})


class BigramCountProgramsTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def convert(self, family, count):
        symbol_text = lambda x: family.tree_distribution_skeleton.symbols[x][0]
        numerators = {
            tuple((symbol_text(sym), pos) for sym, pos in chain): {
                symbol_text(sym): count for sym, count in counts.items()
            }
            for chain, counts in count.numerators.items()
        }
        denominators = {
            tuple((symbol_text(sym), pos) for sym, pos in chain): {
                tuple(symbol_text(sym) for sym in syms): count
                for syms, count in counts.items()
            }
            for chain, counts in count.denominators.items()
        }
        return numerators, denominators

    def count_programs(self, family, programs):
        counts = family.count_programs(
            [[ns.parse_s_expression(x) for x in ps] for ps in programs]
        )
        return [self.convert(family, count) for count in counts.counts]

    def test_counts_single_program(self):
        counts = self.count_programs(fam, [["(+ (1) (2))"]])
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"1": 1},
                        (("+", 1),): {"2": 1},
                    },
                    {
                        # all symbols could be anything else
                        (("<root>", 0),): {("+", "1", "2"): 1},
                        (("+", 0),): {("+", "1", "2"): 1},
                        (("+", 1),): {("+", "1", "2"): 1},
                    },
                )
            ],
        )

    def test_counts_multiple_programs(self):
        counts = self.count_programs(fam, [["(+ (1) (2))", "(+ (1) (1))"]])
        print(counts)

        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 2},
                        (("+", 0),): {"1": 2},
                        (("+", 1),): {"2": 1, "1": 1},
                    },
                    {
                        (("<root>", 0),): {("+", "1", "2"): 2},
                        (("+", 0),): {("+", "1", "2"): 2},
                        (("+", 1),): {("+", "1", "2"): 2},
                    },
                )
            ],
        )

    def test_counts_separate_programs(self):
        counts = self.count_programs(fam, [["(+ (1) (2))"], ["(+ (1) (1))"]])
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"1": 1},
                        (("+", 1),): {"2": 1},
                    },
                    {
                        (("<root>", 0),): {("+", "1", "2"): 1},
                        (("+", 0),): {("+", "1", "2"): 1},
                        (("+", 1),): {("+", "1", "2"): 1},
                    },
                ),
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"1": 1},
                        (("+", 1),): {"1": 1},
                    },
                    {
                        (("<root>", 0),): {("+", "1", "2"): 1},
                        (("+", 0),): {("+", "1", "2"): 1},
                        (("+", 1),): {("+", "1", "2"): 1},
                    },
                ),
            ],
        )

    def test_counts_variables(self):
        counts = self.count_programs(fam_with_vars, [["(call (lam ($0_0)) (1))"]])
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"call": 1},
                        (("call", 0),): {"lam": 1},
                        (("lam", 0),): {"$0_0": 1},
                        (("call", 1),): {"1": 1},
                    },
                    {
                        (("<root>", 0),): {("+", "1", "2", "call"): 1},
                        (("call", 0),): {("lam",): 1},
                        (("lam", 0),): {("$0_0", "+", "1", "2", "call"): 1},
                        (("call", 1),): {("+", "1", "2", "call"): 1},
                    },
                )
            ],
        )

    def test_counts_single_program_ordered(self):
        counts = self.count_programs(fam_with_ordering, [["(+ (1) (2) (3))"]])
        print(counts)
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"1": 1},
                        (("+", 1),): {"2": 1},
                        (("+", 2),): {"3": 1},
                    },
                    {
                        # note that there's no multiplicity here
                        # because the ordering mask doesn't allow it
                        (("<root>", 0),): {("+",): 1},
                        (("+", 0),): {("1",): 1},
                        (("+", 1),): {("2",): 1},
                        (("+", 2),): {("3",): 1},
                    },
                )
            ],
        )

    def test_counts_single_program_ordered_231(self):
        counts = self.count_programs(fam_with_ordering_231, [["(+ (2) (3) (1))"]])
        print(counts)
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"2": 1},
                        (("+", 1),): {"3": 1},
                        (("+", 2),): {"1": 1},
                    },
                    {
                        # note that there's no multiplicity here
                        # because the ordering mask doesn't allow it
                        (("<root>", 0),): {("+",): 1},
                        (("+", 0),): {("2",): 1},
                        (("+", 1),): {("3",): 1},
                        (("+", 2),): {("1",): 1},
                    },
                )
            ],
        )


class BigramParameterDifferenceLossTest(unittest.TestCase):
    def computeLoss(self, logits, programs, family=fam):
        programs = [[ns.parse_s_expression(x) for x in ps] for ps in programs]
        print(programs)
        loss = family.parameter_difference_loss(
            logits,
            family.count_programs(programs),
        )
        loss = loss.detach().cpu().numpy()
        return loss

    def assertLoss(self, logits, programs, target, **kwargs):
        np.testing.assert_almost_equal(
            self.computeLoss(logits, programs, **kwargs), target
        )

    def assertUniformLoss(self, programs, target, **kwargs):
        logits = torch.zeros((1, 4, 2, 4))
        self.assertLoss(logits, programs, target, **kwargs)
        # fill in impossible location +, 0 -> <root>
        logits[0, 1, 0, 0] = 20
        self.assertLoss(logits, programs, target, **kwargs)

    def test_leaf_program_logit(self):
        self.assertUniformLoss([["(1)"]], [np.log(3)])
        self.assertUniformLoss([["(2)"]], [np.log(3)])

    def test_multiple_program_logit(self):
        self.assertUniformLoss([["(1)", "(2)"]], [np.log(3) * 2])
        self.assertUniformLoss([["(1)", "(1)"]], [np.log(3) * 2])

    def test_separate_program_logit(self):
        self.assertUniformLoss([["(1)"], ["(2)"]], [np.log(3)] * 2)
        self.assertUniformLoss([["(1)"], ["(1)"]], [np.log(3)] * 2)

    def test_nested_program_logit(self):
        self.assertUniformLoss([["(+ (1) (2))"]], [np.log(3) * 3])

    def test_nonuniform_loss(self):
        logits = torch.zeros((2, 4, 2, 4))
        logits[0, 0, 0, 1] = np.log(3)  # 1/2 chance of +
        logits[0, 0, 0, 2] = np.log(2)  # 1/3 chance of 1, 1/6 chance of 2
        self.assertLoss(logits, [["(1)"], ["(1)"]], [-np.log(1 / 3), -np.log(1 / 3)])
        self.assertLoss(logits, [["(2)"], ["(2)"]], [-np.log(1 / 6), -np.log(1 / 3)])
        self.assertLoss(
            logits,
            [["(+ (1) (2))"], ["(+ (1) (2))"]],
            [-(np.log(1 / 2) + np.log(1 / 3) * 2), -3 * np.log(1 / 3)],
        )

    def test_variables_loss(self):
        logits = torch.zeros((1, 10, 2, 10))
        self.assertLoss(logits, [["(1)"]], [np.log(4)], family=fam_with_vars)
        self.assertLoss(
            logits,
            [["(call (lam ($0_0)) (1))"]],
            [np.log(4 * 5 * 4)],
            family=fam_with_vars,
        )


class BigramLikelihoodTest(unittest.TestCase):
    def assertLikelihood(self, dist, program, str_prob, family=fam, **kwargs):
        likelihood = family.compute_likelihood(
            dist, ns.parse_s_expression(program), **kwargs
        )
        self.assertEqual(self.render_likelihood(likelihood), str_prob)

    def assertLikelihoods(self, dist, program, nodes_and_probs, family=fam):
        likelihoods = family.compute_likelihood_per_node(
            dist, ns.parse_s_expression(program)
        )
        likelihoods = [
            (ns.render_s_expression(node), self.render_likelihood(prob))
            for node, prob in likelihoods
        ]
        print(likelihoods)
        self.assertEqual(likelihoods, nodes_and_probs)

    def render_likelihood(self, likelihood):
        prob = np.exp(likelihood)
        prob = Fraction.from_float(float(prob)).limit_denominator(1000)
        result = f"log({prob})"
        return result

    def test_leaf(self):
        self.assertLikelihood(fam.uniform(), "(1)", "log(1/3)")

    def test_plus(self):
        self.assertLikelihood(fam.uniform(), "(+ (1) (2))", "log(1/27)")

    def test_plus_nested(self):
        self.assertLikelihood(fam.uniform(), "(+ (1) (+ (1) (2)))", "log(1/243)")
        self.assertLikelihoods(
            fam.uniform(),
            "(+ (1) (+ (1) (2)))",
            [
                ("(+ (1) (+ (1) (2)))", "log(1/3)"),
                ("(1)", "log(1/3)"),
                ("(+ (1) (2))", "log(1/3)"),
                ("(1)", "log(1/3)"),
                ("(2)", "log(1/3)"),
            ],
        )

    def test_leaf_with_variables(self):
        self.assertLikelihood(
            fam_with_vars.uniform(), "(1)", "log(1/4)", family=fam_with_vars
        )

    def test_call_with_variables(self):
        # 1/4 for call
        # 1 for lam
        # 1/5 for $0_0
        # 1/4 for 1
        self.assertLikelihood(
            fam_with_vars.uniform(),
            "(call (lam ($0_0)) (1))",
            "log(1/80)",
            family=fam_with_vars,
        )
        self.assertLikelihoods(
            fam_with_vars.uniform(),
            "(call (lam ($0_0)) (1))",
            [
                ("(call (lam ($0_0)) (1))", "log(1/4)"),
                ("(lam ($0_0))", "log(1)"),
                ("($0_0)", "log(1/5)"),
                ("(1)", "log(1/4)"),
            ],
            family=fam_with_vars,
        )

    def test_ordered(self):
        self.assertLikelihood(
            fam_with_ordering.uniform(),
            "(+ (1) (2) (3))",
            "log(1)",
            family=fam_with_ordering,
        )
        self.assertLikelihood(
            fam_with_ordering.uniform(),
            "(+ (2) (3) (1))",
            "log(0)",
            family=fam_with_ordering,
        )

    def test_ordered_tracker(self):
        def tracker(node, likelihood):
            tracked.append(
                (ns.render_s_expression(node), self.render_likelihood(likelihood))
            )

        tracked = []
        self.assertLikelihood(
            fam_with_ordering.uniform(),
            "(+ (1) (2) (3))",
            "log(1)",
            family=fam_with_ordering,
            tracker=tracker,
        )
        print(tracked)
        self.assertEqual(
            tracked,
            [
                ("(+ (1) (2) (3))", "log(1)"),
                ("(1)", "log(1)"),
                ("(2)", "log(1)"),
                ("(3)", "log(1)"),
            ],
        )

        tracked = []
        self.assertLikelihood(
            fam_with_ordering.uniform(),
            "(+ (2) (3) (1))",
            "log(0)",
            family=fam_with_ordering,
            tracker=tracker,
        )
        print(tracked)
        self.assertEqual(
            tracked,
            [
                ("(+ (2) (3) (1))", "log(1)"),
                ("(2)", "log(0)"),
                ("(3)", "log(0)"),
                ("(1)", "log(0)"),
            ],
        )

    def test_ordered_asserting(self):
        self.assertLikelihood(
            fam_with_ordering_asserted.uniform(),
            "(+ (1) (2) (3))",
            "log(1/27)",
            family=fam_with_ordering_asserted,
        )
        # distribution where 1 is not allowed
        dist = fam_with_ordering_asserted.uniform()
        dist.distribution[:, :, 2] = 0
        self.assertLikelihood(
            dist,
            "(+ (1) (2) (3))",
            "log(0)",
            family=fam_with_ordering_asserted,
        )

    def test_ordered_231(self):
        self.assertLikelihood(
            fam_with_ordering_231.uniform(),
            "(+ (2) (3) (1))",
            "log(1)",
            family=fam_with_ordering_231,
        )
        self.assertLikelihood(
            fam_with_ordering_231.uniform(),
            "(+ (1) (2) (3))",
            "log(0)",
            family=fam_with_ordering_231,
        )

    def test_likelihood_clamped(self, kwargs=lambda x: {}):
        dist = fam.counts_to_distribution(
            fam.count_programs([[ns.parse_s_expression("(1)")]]),
        )[0]
        self.assertEqual(
            fam.compute_likelihood(dist, ns.parse_s_expression("(2)")), -np.inf
        )
        dist = dist.bound_minimum_likelihood(0.01, **kwargs(dist))
        # should be *very* approximately 1/100
        self.assertAlmostEqual(
            fam.compute_likelihood(dist, ns.parse_s_expression("(2)")),
            np.log(1 / 100),
            1,
        )
        # should be *very* approximately 1/100 * 1/3 * 1/3
        # even the (+ (1) (1)) should be since the + -> 1 bigram wasn't seen
        # in the original
        # the 1/3 comes from the fact that in the original the + -> 1 and + -> 2 bigrams
        # were undefined, so now they're uniform
        for code in "(+ (1) (2))", "(+ (1) (1))", "(+ (2) (2))":
            self.assertAlmostEqual(
                fam.compute_likelihood(dist, ns.parse_s_expression(code)),
                np.log(1 / 100 * 1 / 3 * 1 / 3),
                1,
            )

    def test_likelihood_clamped_mask_full(self):
        self.test_likelihood_clamped(
            lambda dist: dict(
                symbol_mask=np.ones((dist.distribution.shape[0]), dtype=np.bool_)
            )
        )

    @parameterized.expand([(True,), (False,)])
    def test_likelihood_clamped_mask_partial(self, include_type_mask):
        family = fam_with_3 if include_type_mask else fam_with_3_no_type_mask
        dist = family.counts_to_distribution(
            family.count_programs([[ns.parse_s_expression("(1)")]]),
        )[0]
        self.assertEqual(
            family.compute_likelihood(dist, ns.parse_s_expression("(2)")), -np.inf
        )
        self.assertEqual(
            [x for x, _ in family.tree_distribution_skeleton.symbols],
            ["<root>", "+", "1", "2", "3"],
        )
        print(np.array(np.where(dist.distribution)).T)
        dist = dist.bound_minimum_likelihood(
            0.01, symbol_mask=np.array([True, True, True, True, False])
        )
        print(np.array(np.where(dist.distribution)).T)
        # should be *very* approximately 1/100
        self.assertAlmostEqual(
            family.compute_likelihood(dist, ns.parse_s_expression("(2)")),
            np.log(1 / 100),
            1,
        )
        # should be *very* approximately 1/100
        self.assertAlmostEqual(
            family.compute_likelihood(dist, ns.parse_s_expression("(3)")),
            -np.inf,
            1,
        )

        # see test_likelihood_clamped for the math
        for code in "(+ (1) (2))", "(+ (1) (1))", "(+ (2) (2))":
            self.assertAlmostEqual(
                family.compute_likelihood(dist, ns.parse_s_expression(code)),
                np.log(1 / 100 * 1 / 3 * 1 / 3),
                1,
            )

        # remains -inf because 3 is impossible
        for code in "(+ (1) (3))", "(+ (3) (3))", "(+ (3) (2))":
            self.assertAlmostEqual(
                family.compute_likelihood(dist, ns.parse_s_expression(code)),
                -np.inf,
                1,
            )


class BigramMixTest(unittest.TestCase):
    def fit_family_to_programs(self, family, programs):
        return family.counts_to_distribution(
            family.count_programs([[ns.parse_s_expression(x) for x in programs]]),
        )[0]

    def merged(self, fam_1, programs_1, fam_2, programs_2, weight):
        dist_1 = self.fit_family_to_programs(fam_1, programs_1)
        dist_2 = self.fit_family_to_programs(fam_2, programs_2)
        return dist_1.mix_with_other(dist_2, weight)

    def assertClose(self, dist_merged, dist_expected):
        delta = dist_merged.distribution - dist_expected.distribution
        bad = np.abs(delta) > 1e-6
        if bad.any():
            syms = dist_merged.dist_fam.symbols()
            for parent, which in np.array(np.where(bad.any(-1))).T:
                actual = dist_merged.distribution[parent, which]
                expected = dist_expected.distribution[parent, which]
                select = (actual > 0) | (expected > 0)
                actual, expected = [
                    {syms[i]: arr[i] for i in np.where(select)[0]}
                    for arr in (actual, expected)
                ]
                self.assertEqual(
                    actual, expected, f"At {syms[parent]}'s {which}th child"
                )

    def assertMix(self, fam_1, programs_1, fam_2, programs_2, weight_1, weight_2):
        self.assertClose(
            self.merged(
                fam_1, programs_1, fam_2, programs_2, weight_2 / (weight_1 + weight_2)
            ),
            self.fit_family_to_programs(
                fam_1 if len(fam_1.symbols()) > len(fam_2.symbols()) else fam_2,
                programs_1 * weight_1 + programs_2 * weight_2,
            ),
        )

    def assertMixBoth(self, fam_1, programs_1, fam_2, programs_2, weight_1, weight_2):
        # pylint: disable=arguments-out-of-order
        self.assertMix(fam_1, programs_1, fam_2, programs_2, weight_1, weight_2)
        self.assertMix(fam_2, programs_2, fam_1, programs_1, weight_2, weight_1)

    progs_1 = "(1)", "(+ (1) (2))"
    progs_2 = "(+ (2) (2))", "(2)"
    progs_3 = "(+ (2) (3))", "(3)"

    def test_merge_bigrams_same_dsl_equal_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam, self.progs_2, 1, 1)

    def test_merge_bigrams_same_dsl_more_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam, self.progs_2, 2, 1)

    def test_merge_bigrams_same_dsl_less_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam, self.progs_2, 2, 5)

    def test_merge_bigrams_with_different_fams_equal_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam_with_3, self.progs_3, 1, 1)

    def test_merge_bigrams_with_different_fams_more_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam_with_3, self.progs_3, 9, 4)

    def test_merge_bigrams_with_different_fams_less_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam_with_3, self.progs_3, 2, 4)

    def test_non_subset_fails(self):
        with self.assertRaises(ValueError) as cm:
            self.merged(fam_with_3, self.progs_3, fam_with_vars, self.progs_1, 1)

        self.assertEqual(
            str(cm.exception),
            "DSL not compatible, extra symbols in this: '3', "
            "extra symbols in other: '$0_0', '$1_0', '$2_0', '$3_0', 'call', 'lam'",
        )


class OrderingTest(unittest.TestCase):
    def test_traversal(self):
        ordering = fam_with_ordering.tree_distribution_skeleton.ordering
        self.assertEqual(
            [
                ns.render_s_expression(x)
                for x in ordering.traverse_preorder(
                    ns.parse_s_expression("(+ (1) (2) (3))")
                )
            ],
            ["(+ (1) (2) (3))", "(1)", "(2)", "(3)"],
        )

    def test_traversal_231(self):
        ordering = fam_with_ordering_231.tree_distribution_skeleton.ordering
        self.assertEqual(
            [
                ns.render_s_expression(x)
                for x in ordering.traverse_preorder(
                    ns.parse_s_expression("(+ (2) (3) (1))")
                )
            ],
            ["(+ (2) (3) (1))", "(1)", "(2)", "(3)"],
        )
        self.assertEqual(
            [
                ns.render_s_expression(x)
                for x in ordering.traverse_preorder(
                    ns.parse_s_expression("(+ (1) (2) (3))")
                )
            ],
            ["(+ (1) (2) (3))", "(3)", "(1)", "(2)"],
        )



import itertools
from abc import ABC
from typing import Callable

from neurosym.search_graph.dsl_search_node import DSLSearchNode
from neurosym.search_graph.metadata_computer import MetadataComputer
from neurosym.types.type_with_environment import Environment, TypeWithEnvironment

from ..dsl.dsl import DSL
from ..programs.hole import Hole, all_holes, replace_holes
from ..programs.s_expression import SExpression
from ..types.type import Type
from .hole_set_chooser import HoleSetChooser
from .search_graph import SearchGraph


class DSLSearchGraph(SearchGraph, ABC):
    """
    Represents a search graph where nodes are SExpressions with holes in them, and edges
    are expansions of those holes.

    :param dsl: DSL to use for expanding holes
    :param target_type: Type of the SExpressions we are searching for
    """

    def __init__(
        self,
        dsl: DSL,
        target_type: Type,
        hole_set_chooser: HoleSetChooser,
        test_predicate: Callable[[SExpression], bool],
        metadata_computer: MetadataComputer,
    ):
        self.dsl = dsl
        self.target_type = target_type
        self.hole_set_chooser = hole_set_chooser
        self.test_predicate = test_predicate
        self.metadata_computer = metadata_computer

    def initial_node(self):
        return DSLSearchNode(
            Hole.of(TypeWithEnvironment(self.target_type, Environment.empty())),
            self.metadata_computer.for_initial_node(),
        )

    def expand_node(self, node):
        hole_sets = self.hole_set_chooser.choose_hole_sets(node.program)
        relevant_holes = sorted(set(h for hs in hole_sets for h in hs))
        # relevant_productions[hole_idx] : list of expansions for holes[hole_idx]
        relevant_productions = {}
        for hole in relevant_holes:
            relevant_productions[hole] = self.dsl.expansions_for_type(hole.twe)
        for hole_set in hole_sets:
            hole_set = sorted(hole_set)
            for hole_replacements in itertools.product(
                *[relevant_productions[h] for h in hole_set]
            ):
                # hole_replacements : list of SExpressions, of length len(holes)
                expanded_program = replace_holes(
                    node.program, hole_set, hole_replacements
                )
                yield DSLSearchNode(
                    expanded_program,
                    self.metadata_computer.for_expanded_node(node, expanded_program),
                )

    def is_goal_node(self, node):
        if any(True for _ in all_holes(node.program)):
            return False
        return self.test_predicate(node)



import unittest
from fractions import Fraction

import numpy as np

import neurosym as ns

from .utils import fit_to


class TestLikelihoodFittedDSL(unittest.TestCase):
    def compute_likelihood(self, corpus, program):
        dfa, _, fam, dist = fit_to(corpus)
        program = ns.to_type_annotated_ns_s_exp(
            ns.python_to_python_ast(program), dfa, "M"
        )
        like = fam.compute_likelihood(dist, program)
        like = Fraction.from_float(float(np.exp(like))).limit_denominator()
        results = fam.compute_likelihood_per_node(dist, program)
        results = [
            (
                ns.render_s_expression(x),
                Fraction.from_float(float(np.exp(y))).limit_denominator(),
            )
            for x, y in results
            if y != 0  # remove zero log-likelihoods
        ]
        print(like)
        print(results)
        return like, results

    def test_likelihood(self):
        like, results = self.compute_likelihood(["x = 2", "y = 3", "y = 4"], "y = 4")
        self.assertAlmostEqual(like, Fraction(2, 9))
        self.assertEqual(
            results,
            [
                ("(const-&y:0~Name)", Fraction(2, 3)),
                ("(const-i4~Const)", Fraction(1, 3)),
            ],
        )

    def test_likelihood_def_use_check(self):
        like, results = self.compute_likelihood(
            ["x = 2; y = x", "y = 2; x = y"], "x = 2; y = x"
        )
        self.assertAlmostEqual(like, Fraction(1, 8))
        self.assertEqual(
            results,
            [
                ("(const-&x:0~Name)", Fraction(1, 2)),
                ("(const-&y:0~Name)", Fraction(1, 2)),
                ("(Name~E (const-&x:0~Name) (Load~Ctx))", Fraction(1, 2)),
            ],
        )

    def test_likelihood_zero(self):
        like, results = self.compute_likelihood(
            ["y = x + 2", "y = 2 + 3", "y = 4"], "y = 2 + x"
        )
        self.assertAlmostEqual(like, Fraction(0))
        self.assertEqual(
            results,
            [
                (
                    ns.render_s_expression(
                        ns.parse_s_expression(
                            """
                            (BinOp~E
                                (Constant~E (const-i2~Const) (const-None~ConstKind))
                                (Add~O) (Name~E (const-g_x~Name) (Load~Ctx)))
                            """
                        )
                    ),
                    Fraction(2, 3),
                ),
                (
                    "(Constant~E (const-i2~Const) (const-None~ConstKind))",
                    Fraction(1, 2),
                ),
                ("(const-i2~Const)", Fraction(1, 2)),
                ("(Name~E (const-g_x~Name) (Load~Ctx))", Fraction(0, 1)),
            ],
        )



from dataclasses import dataclass

import torch
from sklearn.metrics import f1_score, hamming_loss
from torch import nn

from .base_trainer import BaseTrainer, BaseTrainerConfig


@dataclass
class NEARTrainerConfig(BaseTrainerConfig):
    max_seq_len: int = 100
    loss_fn: str = "CrossEntropyLoss"
    num_labels: int = -1  # Set Programmatically


class NEARTrainer(BaseTrainer):
    """
    An abstract class that defines the basic functions to
    implement and train a neural module.
    """

    def __init__(self, model: nn.Module, config: NEARTrainerConfig):
        super().__init__(model=model, config=config)
        assert config.num_labels > 0, "Number of labels must be set programmatically"
        match self.config.loss_fn:
            case "CrossEntropyLoss":
                self.loss_fn = nn.CrossEntropyLoss()
            case "BCEWithLogitsLoss":
                self.loss_fn = nn.BCEWithLogitsLoss()
            case "MSELoss":
                self.loss_fn = nn.MSELoss()
            case "MSELossRegression":
                self.loss_fn = nn.MSELoss()
            case "SmoothL1LossRegression":
                self.loss_fn = nn.SmoothL1Loss()
            case "NLLLoss":
                self.loss_fn = nn.NLLLoss()
            case _:
                raise NotImplementedError(
                    f"Loss function {self.config.loss_fn} not implemented"
                )  # noqa: E501

    @staticmethod
    def compute_average_f1_score(
        predictions: torch.Tensor, targets: torch.Tensor, num_labels: int
    ):  # noqa: E501
        if num_labels > 1:
            weighted_avg_f1 = 1 - f1_score(targets, predictions, average="weighted")
            unweighted_avg_f1 = 1 - f1_score(targets, predictions, average="macro")
            all_f1 = 1 - f1_score(targets, predictions, average=None)
            return dict(
                weighted_avg_f1=weighted_avg_f1,
                unweighted_avg_f1=unweighted_avg_f1,
                all_f1s=all_f1,
            )
        avg_f1 = 1 - f1_score(targets, predictions, average="binary")
        all_f1 = 1 - f1_score(targets, predictions, average=None)
        return dict(avg_f1=avg_f1, all_f1s=all_f1)

    @staticmethod
    def label_correctness(
        predictions: torch.Tensor, targets: torch.Tensor, num_labels: int
    ):  # noqa: E501
        hamming_accuracy = 1 - hamming_loss(
            targets.squeeze().cpu(), predictions.squeeze().cpu()
        )
        f1_scores = NEARTrainer.compute_average_f1_score(
            predictions, targets, num_labels
        )  # noqa: E501
        return dict(hamming_accuracy=hamming_accuracy, **f1_scores)

    def loss(self, predictions: torch.Tensor, targets: torch.Tensor) -> dict:
        # pylint: disable=arguments-differ
        if len(predictions.shape) == 3:
            # Handling seq2seq classification loss.
            match self.config.loss_fn:
                case "CrossEntropyLoss":
                    targets = targets.squeeze(-1)
                    assert (
                        len(targets.shape) == 2
                    ), "Targets must be 2D for classification"
                    predictions = predictions.view(-1, predictions.shape[-1])
                    targets = targets.view(-1)
                case "MSELoss":
                    # pylint: disable=not-callable
                    targets = torch.nn.functional.one_hot(
                        targets.squeeze(-1), num_classes=self.config.num_labels
                    ).float()
                    predictions = predictions.view(-1, predictions.shape[-1])
                    targets = targets.view(-1, targets.shape[-1])
                case "MSELossRegression":
                    # pylint: disable=not-callable
                    predictions = predictions.view(-1, predictions.shape[-1])
                    targets = targets.view(-1, targets.shape[-1])
                case "SmoothL1LossRegression":
                    # pylint: disable=not-callable
                    predictions = predictions.view(-1, predictions.shape[-1])
                    targets = targets.view(-1, targets.shape[-1])
                case "NLLLoss":
                    predictions = (
                        predictions.view(-1, predictions.shape[-1])
                        .clamp(min=1e-10)
                        .log()
                        .log_softmax(dim=-1)
                    )
                    targets = targets.view(-1)
                case _:
                    raise NotImplementedError(
                        f"Loss function {self.config.loss_fn} not implemented for seq2seq models"
                    )
        loss = self.loss_fn(predictions, targets)
        return loss

    def _step(self, inputs, outputs, validation=False, **kwargs):
        # pylint: disable=arguments-differ
        del kwargs
        predictions = self.model(inputs)
        losses = dict(loss=self.loss(predictions, outputs))

        if validation and self.logger is not None and self.current_epoch % 2 == 0:
            self._log_program_accuracy(predictions, outputs, inputs)

        return losses

    def _log_program_accuracy(self, predictions, outputs, inputs):
        del inputs
        if self.config.num_labels > 1:
            predictions = torch.argmax(predictions, dim=-1)
        else:
            predictions = torch.round(torch.sigmoid(predictions))
        targets = outputs
        correctness = NEARTrainer.label_correctness(
            predictions, targets, self.config.num_labels
        )
        self.logger.log_metrics(correctness, step=self.global_step)


def main():
    from neurosym.utils.imports import import_pytorch_lightning

    pl = import_pytorch_lightning()
    from neurosym.datasets.load_data import DatasetFromNpy, DatasetWrapper

    dataset_factory = lambda train_seed: DatasetWrapper(
        DatasetFromNpy(
            "../data/classification_example/train_ex_data.npy",
            "../data/classification_example/train_ex_labels.npy",
            train_seed,
        ),
        DatasetFromNpy(
            "../data/classification_example/test_ex_data.npy",
            "../data/classification_example/test_ex_labels.npy",
            None,
        ),
        batch_size=200,
    )
    datamodule = dataset_factory(42)
    # model = nn.Sequential(
    #     nn.Linear(2, 100),
    #     nn.ReLU(),
    #     nn.Linear(100, 2),
    # )
    model = nn.Linear(2, 2)
    # model.weight.data = torch.tensor([[0., 1.], [0., 0.]])
    # model.bias.data = torch.tensor([0., 0.])
    trainer_cfg = NEARTrainerConfig(
        lr=1e-4,
        max_seq_len=100,
        n_epochs=10000,
        num_labels=2,
        train_steps=len(datamodule.train),
        loss_fn="NLLLoss",
    )
    pl_model = NEARTrainer(model, config=trainer_cfg)
    trainer = pl.Trainer(
        max_epochs=trainer_cfg.n_epochs,
        devices="auto",
        accelerator="cpu",
        enable_checkpointing=False,
        enable_model_summary=False,
        logger=False,
        callbacks=[],
    )
    trainer.fit(pl_model, datamodule.train_dataloader(), datamodule.val_dataloader())
    trainer.validate(pl_model, datamodule.val_dataloader())

    print(trainer.callback_metrics)


if __name__ == "__main__":
    main()



import ast
import json
from functools import lru_cache
from textwrap import dedent

import neurosym as ns


def fit_to(
    programs,
    parser=ns.python_to_python_ast,
    root="M",
    use_def_use=True,
    use_node_ordering=True,
    include_type_preorder_mask=True,
):
    """
    Set include_type_preorder_mask to False to disable the type preorder mask,
        this is basically only useful in the specific context where we are testing
        the names mask and want no other masks to be applied.
    """
    dfa = ns.python_dfa()
    programs = [parser(p) for p in programs]
    subset = ns.PythonDSLSubset.from_programs(dfa, *programs, root=root)
    dsl = ns.create_python_dsl(dfa, subset, root)
    apms = [
        lambda dist, dsl: ns.python_def_use_mask.DefUseChainPreorderMask(
            dist, dsl, ns.python_def_use_mask.DefUseMaskConfiguration(dfa, {})
        )
    ]
    node_ordering = (
        ns.python_def_use_mask.PythonNodeOrdering
        if use_node_ordering
        else ns.DefaultNodeOrdering
    )
    fam = ns.BigramProgramDistributionFamily(
        dsl,
        additional_preorder_masks=apms if use_def_use else [],
        include_type_preorder_mask=include_type_preorder_mask,
        node_ordering=node_ordering,
    )
    counts = fam.count_programs(
        [[ns.to_type_annotated_ns_s_exp(program, dfa, root) for program in programs]]
    )
    dist = fam.counts_to_distribution(counts)[0]
    return dfa, dsl, fam, dist


@lru_cache(None)
def small_set_runnable_code_examples():
    with open("test_data/small_set_runnable_code.json") as f:
        contents = json.load(f)
    return contents


def cwq(s):
    """
    Canonicalize with question marks and dollars
    """
    s = dedent(s)
    s = s.replace("?", "__QUESTION__MARK__")
    s = s.replace("$", "__DOLLAR__")
    s = ast.unparse(ast.parse(s))
    s = s.replace("__QUESTION__MARK__", "?")
    s = s.replace("__DOLLAR__", "$")
    return s



from abc import ABC, abstractmethod
from typing import List

from ..programs.hole import Hole, all_holes
from ..programs.s_expression import SExpression


class HoleSetChooser(ABC):
    @abstractmethod
    def choose_hole_sets(self, node: SExpression) -> List[List[Hole]]:
        """
        Returns a list of sets of holes, where each set of holes will be expanded
        simultaneously
        """


class ChooseFirst(HoleSetChooser):
    def choose_hole_sets(self, node: SExpression) -> List[List[Hole]]:
        result = []
        for hole in all_holes(node):
            result.append([hole])
            break
        return result


class ChooseAll(HoleSetChooser):
    def choose_hole_sets(self, node: SExpression) -> List[List[Hole]]:
        result = []
        for hole in all_holes(node):
            result.append([hole])
        return result



from collections import defaultdict
from dataclasses import dataclass
from types import NoneType
from typing import Callable, Dict, List, Tuple, Union

import numpy as np
import torch

from neurosym.dsl.dsl import DSL, ROOT_SYMBOL
from neurosym.program_dist.tree_distribution.ordering import DefaultNodeOrdering
from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    ConjunctionPreorderMask,
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.preorder_mask.type_preorder_mask import (
    TypePreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression
from neurosym.types.type import Type

from .tree_distribution.tree_distribution import TreeProgramDistributionFamily


@dataclass
class BigramProgramDistribution:
    dist_fam: "BigramProgramDistributionFamily"
    distribution: np.ndarray

    def __post_init__(self):
        assert self.distribution.ndim == 3
        assert self.distribution.shape[0] == self.distribution.shape[2]

    def bound_minimum_likelihood(
        self, min_likelihood: float, symbol_mask: np.ndarray = None
    ):
        assert 0 <= min_likelihood <= 1
        if symbol_mask is not None:
            assert (
                len(symbol_mask.shape) == 1
                and symbol_mask.shape[0] == self.distribution.shape[0]
            )
            assert symbol_mask.dtype == np.bool_
        distribution = self.distribution
        if symbol_mask is None:
            distribution = np.maximum(distribution, min_likelihood)
        else:
            mask_square = self._square_mask(symbol_mask)
            distribution = distribution.copy()
            distribution[mask_square] = np.maximum(
                distribution[mask_square], min_likelihood
            )
        distribution = self.dist_fam.mask_invalid(distribution)
        distribution = distribution / distribution.sum(-1)[..., None]
        return BigramProgramDistribution(self.dist_fam, distribution)

    def _square_mask(self, symbol_mask):
        mask_square = symbol_mask[:, None] & symbol_mask[None, :]
        mask_square = mask_square[:, None, :].repeat(self.distribution.shape[1], axis=1)

        return mask_square

    def mix_with_other(self, other: "BigramProgramDistribution", weight_other: float):
        # pylint: disable=self-cls-assignment
        assert 0 <= weight_other <= 1
        symbols_this = self.dist_fam.symbols()
        symbols_other = other.dist_fam.symbols()
        if set(symbols_this).issubset(symbols_other):
            self, other = other, self
            symbols_this, symbols_other = symbols_other, symbols_this
            weight_other = 1 - weight_other
        else:
            if not set(symbols_this).issuperset(symbols_other):
                extra_this = set(symbols_this) - set(symbols_other)
                extra_other = set(symbols_other) - set(symbols_this)
                extra_this, extra_other = (
                    ", ".join(repr(x) for x in sorted(extra))
                    for extra in (extra_this, extra_other)
                )
                raise ValueError(
                    "DSL not compatible, extra symbols in this: "
                    f"{extra_this}, extra symbols in other: {extra_other}"
                )
        mask = np.array([s in symbols_other for s in symbols_this], dtype=np.bool_)
        distribution = self.distribution.copy()
        distribution_other = np.zeros_like(distribution)
        distribution_other[self._square_mask(mask)] = other.distribution.flatten()
        distribution[mask] = (
            weight_other * distribution_other[mask]
            + (1 - weight_other) * distribution[mask]
        )
        return BigramProgramDistribution(self.dist_fam, distribution)


@dataclass
class BigramProgramDistributionBatch:
    dist_fam: "BigramProgramDistributionFamily"
    distribution_batch: np.ndarray

    def __post_init__(self):
        assert isinstance(self.distribution_batch, np.ndarray), type(
            self.distribution_batch
        )
        assert self.distribution_batch.ndim == 4
        assert self.distribution_batch.shape[1] == self.distribution_batch.shape[3]

    def __getitem__(self, i):
        return BigramProgramDistribution(self.dist_fam, self.distribution_batch[i])

    def __len__(self):
        return len(self.distribution_batch)


@dataclass
class BigramProgramCounts:
    # map from (parent_sym, parent_child_idx) to map from child_sym to count
    numerators: Dict[Tuple[Tuple[int, int], ...], Dict[int, int]]
    # map from (parent_sym, parent_child_idx) to map from potential child_sym values to count
    denominators: Dict[Tuple[Tuple[int, int], ...], Dict[Tuple[int, ...], int]]

    def add_to_numerator_array(self, arr, batch_idx):
        for [(parent_sym, parent_child_idx)], children in self.numerators.items():
            for child_sym, count in children.items():
                arr[batch_idx, parent_sym, parent_child_idx, child_sym] = count
        return arr

    def add_to_denominator_array(self, arr, batch_idx, backmap):
        for [(parent_sym, parent_child_idx)], children in self.denominators.items():
            for child_syms, count in children.items():
                key = batch_idx, parent_sym, parent_child_idx, backmap[child_syms]
                arr[key] = count
        return arr


@dataclass
class BigramProgramCountsBatch:
    dist_fam: "BigramProgramDistributionFamily"
    counts: List[BigramProgramCounts]

    def numerators(self, num_symbols, max_arity):
        numerators = np.zeros(
            (len(self.counts), num_symbols, max_arity, num_symbols), dtype=np.int32
        )
        for i, dist in enumerate(self.counts):
            dist.add_to_numerator_array(numerators, i)
        return numerators

    def denominators(self, num_symbols, max_arity):
        denominator_keys = {
            key
            for counts in self.counts
            for mapping in counts.denominators.values()
            for key in mapping.keys()
        }
        denominator_keys = sorted(denominator_keys)
        denominator_keys_backmap = {key: i for i, key in enumerate(denominator_keys)}
        denominators = np.zeros(
            (len(self.counts), num_symbols, max_arity, len(denominator_keys)),
            dtype=np.int32,
        )
        for i, counts in enumerate(self.counts):
            counts.add_to_denominator_array(denominators, i, denominator_keys_backmap)

        return denominators, denominator_keys

    def to_distribution(self, num_symbols, max_arity):
        numerators = self.numerators(num_symbols, max_arity)

        # We do not need to handle denominators here, as this is just
        # a simple conversion from counts to probabilities, and we do
        # not need to handle the case where the denominator is 0.

        return BigramProgramDistributionBatch(
            self.dist_fam, counts_to_probabilities(numerators)
        )


class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
    def __init__(
        self,
        dsl: DSL,
        valid_root_types: Union[NoneType, List[Type]] = None,
        *,
        additional_preorder_masks: Tuple[
            Callable[[DSL, TreeDistribution], PreorderMask]
        ] = (),
        include_type_preorder_mask: bool = True,
        node_ordering=DefaultNodeOrdering,
    ):
        if valid_root_types is not None:
            dsl = dsl.with_valid_root_types(valid_root_types)
        self._dsl = dsl
        self._symbols, self._arities, self._valid_mask = bigram_mask(dsl)
        self._max_arity = max(self._arities)
        self._symbol_to_idx = {sym: i for i, sym in enumerate(self._symbols)}
        self._additional_preorder_masks = additional_preorder_masks
        self._include_type_preorder_mask = include_type_preorder_mask
        self._node_ordering = node_ordering

    def underlying_dsl(self) -> DSL:
        return self._dsl

    def parameters_shape(self) -> List[int]:
        return self._valid_mask.shape

    def normalize_parameters(
        self, parameters: torch.Tensor, *, logits: bool, neg_inf=-float("inf")
    ) -> torch.Tensor:
        parameters = parameters.clone()
        mask = torch.tensor(self._valid_mask, device=parameters.device)[None].repeat(
            parameters.shape[0], 1, 1, 1
        )
        parameters[~mask] = -float("inf")
        if logits:
            parameters = parameters.log_softmax(-1)
            parameters[~mask] = neg_inf
        else:
            parameters = parameters.softmax(-1)
            parameters[~mask] = 0
        return parameters

    def with_parameters(
        self, parameters: torch.Tensor
    ) -> BigramProgramDistributionBatch:
        assert (
            parameters.shape[1:] == self.parameters_shape()
        ), f"Expected {self.parameters_shape()}, got {parameters.shape[1:]}"
        parameters = self.normalize_parameters(parameters, logits=False)
        return BigramProgramDistributionBatch(self, parameters.detach().cpu().numpy())

    def count_programs(self, data: List[List[SExpression]]) -> BigramProgramCountsBatch:
        tree_dist = self.tree_distribution_skeleton
        all_counts = []
        for programs in data:
            numerators, denominators = count_programs(tree_dist, programs)
            all_counts.append(
                BigramProgramCounts(numerators=numerators, denominators=denominators)
            )
        return BigramProgramCountsBatch(self, all_counts)

    def counts_to_distribution(
        self, counts: BigramProgramCountsBatch
    ) -> BigramProgramDistribution:
        return counts.to_distribution(len(self._symbols), self._max_arity)

    def parameter_difference_loss(
        self, parameters: torch.tensor, actual: BigramProgramCountsBatch
    ) -> torch.float32:
        """
        Let
            p be a program
            g be a bigram context (i.e. parent +, in position 2)
            s be a symbol
            d be the "denominator" (which is the set of *possible* symbols s'
                that could have appeared in this context g).
        The ngram `g` is really the context that `s` appears in. So for example
            `(+ (- 1 2) 3)` if we're thinking about the symbol s=`-` its bigram g=`(+, 0)`
            indicating it's the first argument of a "+". This is the bigram notion from
            DreamCoder where it's not just the parent (analogous to "previous token" in NLP)
            but also includes which child of the parent we are. In general, we also filter
            on other characteristics (e.g., type checking), not just ngrams. See the
            preorder mask for more details.

        In this particular example, `d` would be the set of symbols that *could* have
            appeared in that same position `g`, for example `1` or `+` or `-` or anything
            else that type checks. This is called the denominator because the
            probability of choosing `s` among `(s' in d)` is going to be
                P(s)/sum_{s' in d} P(s')
            These probabilities will depend on the parameters `theta` of the bigram model
            (and of course, specifically the unigram it assigns to the context `g`).

        (See also math below) Our goal is to compute the loglikelihood of the actual
            program P under the bigram parameters theta. Which, for a bigram is the sum
            of the logprob over all subtrees of the symbol `s` context `g` and denominator `d` for
            that subtree. We can factor this overall sum into two parts: a numerator based on
            sthe actual symbol `s` and a denominator based on the alternative symbols in `d`
            for that context `g`.

        sum_p log P(p | theta)
            = sum_p sum_{(g, s, d) in p} log P(s | g, theta, d)
            = sum_p sum_{(g, s, d) in p} log (exp(theta_{g, s}) / sum_{s' in d} exp(theta_{g, s'}))
            = sum_p sum_{(g, s, d) in p} (theta_{g, s} - log sum_{s' in d} exp(theta_{g, s'}))
            = [sum_p sum_{(g, s, d) in p} theta_{g, s}]
                - [sum_p sum_{(g, s, d) in p} log sum_{s' in d} exp(theta_{g, s'})]
            = [numer] - [denom]

        For each (g,s) instance in the corpus the numerator is the same, and for each (g,d)
            instance the denominator is the same, so we can instead come up with counts for
            each of these (calling them `numcount` and `dencount`) and rewrite our sum as a
            sum over the unique (g,s) and (g,d) instances:

        numer
            = sum_p sum_{(g, s, d) in p} theta_{g, s}
            = sum_g sum_s numcount_{g, s} theta_{g, s}
            = (numcount * theta).sum()

        denom
            = sum_p sum_{(g, s, d) in p} log sum_{s' in d} exp(theta_{g, s'})
            = sum_g sum_d dencount_{g, d} log sum_{s' in d} exp(theta_{g, s'})

        theta_by_denom(g, s', d) = theta_{g, s'} if s' in d, else -inf

        denom
            = sum_g sum_d dencount_{g, d} log sum_{s'} exp(theta_by_denom{g, s', d})

        agg_theta_by_denom(g, d) = logsumexp(theta_by_denom(g, *, d))

        denom
            = sum_g sum_d dencount_{g, d} agg_theta_by_denom(g, d)
            = (dencount * agg_theta_by_denom).sum()

        """

        assert parameters.shape[1:] == self.parameters_shape()
        assert len(parameters.shape) == 4

        numcount = actual.numerators(len(self._symbols), self._max_arity)
        dencount, den_keys = actual.denominators(len(self._symbols), self._max_arity)
        numcount, dencount = [
            torch.tensor(x, device=parameters.device, dtype=torch.float32)
            for x in (numcount, dencount)
        ]

        def agg_across_all_but_batch_axis(x, fn):
            x = x.reshape(x.shape[0], -1)
            return fn(x, -1)

        numer = agg_across_all_but_batch_axis(numcount * parameters, torch.sum)

        theta_by_denom = parameters[..., None].repeat(1, 1, 1, 1, len(den_keys))
        for i, key in enumerate(den_keys):
            mask = torch.ones(
                len(self._symbols), dtype=torch.bool, device=parameters.device
            )
            mask[list(key)] = False
            theta_by_denom[..., mask, i] = -float("inf")
        agg_theta_by_denom = torch.logsumexp(theta_by_denom, dim=-2)
        denom = agg_across_all_but_batch_axis(dencount * agg_theta_by_denom, torch.sum)
        return -(numer - denom)

    def uniform(self):
        return BigramProgramDistribution(
            self, counts_to_probabilities(self._valid_mask)
        )

    def compute_tree_distribution(
        self, distribution: Union[BigramProgramDistribution, NoneType]
    ) -> TreeDistribution:
        if isinstance(distribution, BigramProgramDistribution):
            assert isinstance(distribution, BigramProgramDistribution), type(
                distribution
            )
            dist_vals = distribution.distribution
        else:
            assert distribution is None
            dist_vals = self._valid_mask

        dist = defaultdict(list)
        for parent, position, child in zip(*np.where(dist_vals > 0)):
            dist[(parent, position),].append(
                (child, np.log(dist_vals[parent, position, child]))
            )
        dist = {k: sorted(v, key=lambda x: -x[1]) for k, v in dist.items()}

        return TreeDistribution(
            1,
            dist,
            list(zip(self._symbols, self._arities)),
            self.compute_preorder_mask,
            self._node_ordering,
        )

    def compute_preorder_mask(self, tree_dist):
        masks = []
        if self._include_type_preorder_mask:
            masks.append(TypePreorderMask(tree_dist, self._dsl))
        for mask in self._additional_preorder_masks:
            masks.append(mask(tree_dist, self._dsl))
        return ConjunctionPreorderMask(tree_dist, masks)

    def mask_invalid(self, distribution):
        return distribution * self._valid_mask

    def symbols(self):
        return self._symbols


def bigram_mask(dsl):
    symbols = dsl.ordered_symbols(include_root=True)

    valid_root_types = dsl.valid_root_types
    symbol_to_idx = {sym: i for i, sym in enumerate(symbols)}
    rules_for = dsl.all_rules(
        care_about_variables=False, valid_root_types=valid_root_types
    )
    arities = [None] * len(symbols)
    arities[0] = 1
    for _, rules in rules_for.items():
        for sym, children in rules:
            assert arities[symbol_to_idx[sym]] in (None, len(children)), str(
                symbol_to_idx
            )
            arities[symbol_to_idx[sym]] = len(children)
    assert all(
        x is not None for x in arities
    ), f"No arities for symbols {[sym for sym, ar in zip(symbols, arities) if ar is None]}"
    valid_mask = np.zeros((len(symbols), max(arities), len(symbols)), dtype=np.bool_)
    root_syms = {sym for t in valid_root_types for sym, _ in rules_for[t]}
    for root_sym in root_syms:
        valid_mask[symbol_to_idx[ROOT_SYMBOL], 0, symbol_to_idx[root_sym]] = 1
    for _, rules in rules_for.items():
        for root_sym, child_types in rules:
            for i, ct in enumerate(child_types):
                for child_sym, _ in rules_for.get(ct, ()):
                    valid_mask[symbol_to_idx[root_sym], i, symbol_to_idx[child_sym]] = 1
    return symbols, np.array(arities), valid_mask


def counts_to_probabilities(counts):
    return np.divide(
        counts,
        counts.sum(-1)[..., None],
        out=np.zeros_like(counts, dtype=np.float32),
        where=counts != 0,
    )


def count_programs(tree_dist: TreeDistribution, programs: List[SExpression]):
    """
    Count the productions in the programs, indexed by the path to the node.
    """
    numerators = defaultdict(lambda: defaultdict(int))
    denominators = defaultdict(lambda: defaultdict(int))
    for program in programs:
        preorder_mask = tree_dist.mask_constructor(tree_dist)
        preorder_mask.on_entry(0, 0)
        accumulate_counts(
            tree_dist,
            program,
            numerators,
            denominators,
            ((0, 0),),
            preorder_mask=preorder_mask,
        )
    numerators = {k: dict(v) for k, v in numerators.items()}
    denominators = {k: dict(v) for k, v in denominators.items()}
    return numerators, denominators


def accumulate_counts(
    tree_dist: TreeDistribution,
    program: SExpression,
    numerators: Dict[Tuple[Tuple[int, int], ...], Dict[int, int]],
    denominators: Dict[Tuple[Tuple[int, int], ...], Dict[Tuple[int, ...], int]],
    ancestors: Tuple[Tuple[int, int], ...],
    preorder_mask: PreorderMask,
):
    parent_position = ancestors[-1][1]
    this_idx = tree_dist.symbol_to_index[program.symbol]
    numerators[ancestors][this_idx] += 1
    possibilities = np.arange(len(tree_dist.symbols))
    mask = preorder_mask.compute_mask(parent_position, possibilities)
    preorder_mask.on_entry(parent_position, this_idx)
    elements = possibilities[mask]
    elements = tuple(int(x) for x in elements)
    denominators[ancestors][elements] += 1
    order = tree_dist.ordering.order(this_idx, len(program.children))
    for j, child in zip(order, [program.children[i] for i in order]):
        new_ancestors = ancestors + ((this_idx, j),)
        new_ancestors = new_ancestors[-tree_dist.limit :]
        accumulate_counts(
            tree_dist,
            child,
            numerators,
            denominators,
            new_ancestors,
            preorder_mask=preorder_mask,
        )
    preorder_mask.on_exit(parent_position, this_idx)



from collections import defaultdict
from dataclasses import dataclass, field
from types import NoneType
from typing import Callable, Dict, List, Set, Tuple, Union

from neurosym.dsl.dsl_factory import DSLFactory
from neurosym.programs.s_expression import SExpression
from neurosym.python_dsl import python_ast_tools
from neurosym.python_dsl.convert_python.convert import to_type_annotated_ns_s_exp
from neurosym.python_dsl.convert_python.python_ast import PythonAST
from neurosym.types.type import ArrowType
from neurosym.types.type_string_repr import parse_type, render_type

from .names import PYTHON_DSL_SEPARATOR


@dataclass
class PythonDSLSubset:
    """
    Represents a subset of the python DSL. This is represented as
        - a dictionary from sequential type to a list of lengths.
        - a dictionary from types to a list of leaves of that type
    """

    _lengths_by_sequence_type: Dict[str, Set[int]] = field(
        default_factory=lambda: defaultdict(set)
    )
    _leaves: Dict[str, Set[str]] = field(default_factory=lambda: defaultdict(set))

    @property
    def lengths_by_sequence_type(self) -> Dict[str, List[int]]:
        return {k: sorted(v) for k, v in self._lengths_by_sequence_type.items()}

    @property
    def leaves(self) -> Dict[str, List[str]]:
        return {k: sorted(v) for k, v in self._leaves.items()}

    def add_s_exps(self, *s_exps):
        """
        Add the following s-expressions to the subset. They must be type-annotated.
        """
        for s_exp in s_exps:
            for node in traverse(s_exp):
                assert isinstance(node, SExpression)
                symbol, state, *_ = node.symbol.split(PYTHON_DSL_SEPARATOR)
                state = python_ast_tools.unclean_type(state)
                if python_ast_tools.is_sequence(state, symbol):
                    self._lengths_by_sequence_type[state].add(len(node.children))
                elif len(node.children) == 0:
                    self._leaves[state].add(symbol)

    @classmethod
    def from_s_exps(cls, s_exps):
        """
        Factory version of add_s_exps.
        """
        subset = cls()
        subset.add_s_exps(*s_exps)
        return subset

    def add_programs(
        self,
        dfa,
        *programs: Tuple[PythonAST, ...],
        root: Union[str, Tuple[str, ...]],
    ):
        """
        Add the programs to the subset. The root symbol of the DSL is passed as an argument,
            and can be a single string or a tuple of strings.

        Args:
            dfa: the dfa of the DSL
            programs: the programs to extract the subset from
            root: the root symbol of the DSL. If a tuple is passed, it must
                be the same length as the programs, providing a root symbol for each program.
            abstrs: abstractions: their bodies will be added to the list of programs
        """
        if isinstance(root, str):
            root = [root] * len(programs)
        else:
            if len(root) != len(programs):
                raise ValueError(
                    "The length of the root should be the same as the number of programs, but was"
                    f" {len(root)} and {len(programs)} respectively."
                )

        s_exps = []
        for program, root_sym in zip(programs, root):
            s_exp = to_type_annotated_ns_s_exp(program, dfa, root_sym)
            self.add_s_exps(s_exp)
            s_exps.append(s_exp)
        return s_exps

    @classmethod
    def from_programs(
        cls, dfa, *programs: Tuple[PythonAST, ...], root: Union[str, Tuple[str, ...]]
    ):
        """
        Factory version of add_programs.
        """
        subset = cls()
        subset.add_programs(dfa, *programs, root=root)
        return subset

    def fill_in_missing_lengths(self):
        """
        Fill in "missing lengths" for each sequence type. E.g., if the lengths
            of a sequence type are [1, 3], this function will add 2 to the list.
        """
        self._lengths_by_sequence_type = {
            seq_type: set(range(min(lengths), max(lengths) + 1))
            for seq_type, lengths in self.lengths_by_sequence_type.items()
        }


def traverse(s_exp):
    """
    Yield all the nodes in the s-expression.
    """
    yield s_exp
    for child in s_exp.children:
        yield from traverse(child)


def create_python_dsl(
    dfa: dict,
    dsl_subset: PythonDSLSubset,
    start_state: str,
    add_additional_productions: Callable[[DSLFactory], NoneType] = lambda dslf: None,
):
    """
    Create a DSL from a DFA and a subset of the DSL.

    Args:
        dfa: the DFA of the DSL
        dsl_subset: the subset of the DSL
        start_state: the start state of the DSL
        add_additional_productions: a function that adds additional productions to the DSL

    Returns:
        the DSL
    """
    dslf = DSLFactory()
    for target in dfa:
        for prod in dfa[target]:
            input_types = tuple(parse_type(t) for t in dfa[target][prod])
            if python_ast_tools.is_sequence(target, prod):
                assert len(input_types) == 1
                for length in dsl_subset.lengths_by_sequence_type.get(target, []):
                    typ = ArrowType(input_types * length, parse_type(target))
                    dslf.concrete(
                        prod
                        + PYTHON_DSL_SEPARATOR
                        + python_ast_tools.clean_type(target)
                        + PYTHON_DSL_SEPARATOR
                        + str(length),
                        render_type(typ),
                        None,
                    )
            else:
                typ = ArrowType(tuple(input_types), parse_type(target))
                dslf.concrete(
                    prod + PYTHON_DSL_SEPARATOR + python_ast_tools.clean_type(target),
                    render_type(typ),
                    None,
                )
    for target, leaves in dsl_subset.leaves.items():
        for constant in leaves:
            typ = ArrowType((), parse_type(target))
            dslf.concrete(
                constant + PYTHON_DSL_SEPARATOR + target, render_type(typ), None
            )
    add_additional_productions(dslf)
    dslf.prune_to(start_state, tolerate_pruning_entire_productions=True)
    return dslf.finalize()



from abc import ABC, abstractmethod
from dataclasses import KW_ONLY, dataclass
from typing import Callable, Dict

from ..types.type_signature import (
    LambdaTypeSignature,
    TypeSignature,
    VariableTypeSignature,
)


class Production(ABC):
    """
    Represents a production rule in a simple s-expression grammar.

    Has a symbol, a type signature, and a function that represents the fn represented.
    """

    @abstractmethod
    def base_symbol(self):
        pass

    @abstractmethod
    def get_index(self):
        pass

    @abstractmethod
    def with_index(self, index):
        pass

    def symbol(self):
        """
        Return the symbol of this production.
        """
        if self.get_index() is None:
            return self.base_symbol()
        return f"{self.base_symbol()}_{self.get_index()}"

    @abstractmethod
    def type_signature(self) -> TypeSignature:
        """
        Return a TypeSignature object representing the type signature of this production
        """

    @abstractmethod
    def initialize(self, dsl) -> Dict[str, object]:
        """
        Return some state that this production might need to compute its function.
        E.g., for a neural network production, this might be the weights of the network.
        """

    @abstractmethod
    def apply(self, dsl, state, children):
        """
        Apply this production to the given children.
        """

    @abstractmethod
    def render(self) -> str:
        """
        Render this production as a string.
        """

    @classmethod
    def reindex(cls, productions):
        """
        Reindex the given productions, so that they are indexed from 0 to len(productions) - 1.
        """
        if len(productions) == 1:
            return [productions[0].with_index(None)]
        return [p.with_index(i) for i, p in enumerate(productions)]


class FunctionLikeProduction(Production):
    @abstractmethod
    def evaluate(self, dsl, state, inputs):
        """
        Return the resulting pytorch expression of computing this function on the inputs
            Takes in the state of the production, which is the result of initialize().

        Effectively a form of denotation semantics.
        """

    def apply(self, dsl, state, children):
        """
        Apply this production to the given children.
        """
        return self.evaluate(dsl, state, [dsl.compute(x) for x in children])


@dataclass
class ConcreteProduction(FunctionLikeProduction):
    _symbol: str
    _type_signature: TypeSignature
    _compute: Callable[..., object]
    _: KW_ONLY
    index: int = None

    def base_symbol(self):
        return self._symbol

    def get_index(self):
        return self.index

    def with_index(self, index):
        return type(self)(
            self._symbol, self._type_signature, self._compute, index=index
        )

    def type_signature(self) -> TypeSignature:
        return self._type_signature

    def initialize(self, dsl) -> Dict[str, object]:
        del dsl
        return {}

    def evaluate(self, dsl, state, inputs):
        del dsl
        assert state == {}
        try:
            return self._compute(*inputs)
        except TypeError as exc:
            raise TypeError(
                f"Error computing {self.symbol()} on inputs {inputs} with state {state}"
            ) from exc

    def render(self):
        return f"{self.symbol():>15} :: {self._type_signature.render()}"


@dataclass
class LambdaProduction(Production):
    _unique_id: int
    _type_signature: LambdaTypeSignature

    @property
    def arity(self):
        return self._type_signature.function_arity()

    def base_symbol(self):
        return "lam"

    def get_index(self):
        return self._unique_id

    def with_index(self, index):
        return LambdaProduction(index, self._type_signature)

    def type_signature(self) -> TypeSignature:
        return self._type_signature

    def initialize(self, dsl) -> Dict[str, object]:
        del dsl
        return {}

    def apply(self, dsl, state, children):
        """
        Apply this production to the given children.
        """
        # pylint: disable=cyclic-import
        from neurosym.dsl.lambdas import LambdaFunction

        [body] = children
        return LambdaFunction.of(dsl, body, self._type_signature)

    def render(self):
        return f"{self.symbol():>15} :: {self._type_signature.render()}"


@dataclass
class VariableProduction(Production):
    _unique_id: int
    _type_signature: VariableTypeSignature

    def base_symbol(self):
        return f"${self._type_signature.index_in_env}"

    def get_index(self):
        return self._unique_id

    def with_index(self, index):
        return VariableProduction(index, self._type_signature)

    def type_signature(self) -> TypeSignature:
        return self._type_signature

    def initialize(self, dsl) -> Dict[str, object]:
        del dsl
        return {}

    def apply(self, dsl, state, children):
        """
        Apply this production to the given children.
        """
        raise NotImplementedError

    def render(self):
        return f"{self.symbol():>15} :: {self._type_signature.render()}"

    @property
    def index_in_env(self):
        return self._type_signature.index_in_env


@dataclass
class ParameterizedProduction(ConcreteProduction):
    initializers: Dict[str, Callable[[], object]]

    def with_index(self, index):
        # pylint: disable=unexpected-keyword-arg
        return ParameterizedProduction(
            self._symbol,
            self._type_signature,
            self._compute,
            index=index,
            initializers=self.initializers,
        )

    def initialize(self, dsl) -> Dict[str, object]:
        del dsl
        return {k: v() for k, v in self.initializers.items()}

    def evaluate(self, dsl, state, inputs):
        del dsl
        return self._compute(*inputs, **state)

    def render(self):
        lhs = f"{self.symbol()}[{', '.join(self.initializers)}]"
        return f"{lhs:>15} :: {self._type_signature.render()}"



import ast
import uuid
from dataclasses import dataclass

import ast_scope
from no_toplevel_code import wrap_ast

from neurosym.python_dsl import python_ast_tools


@dataclass(frozen=True)
class PythonSymbol:
    """
    Represents a symbol, like &x:3. This means the symbol x in static frame 3.
    Can also represent a global symbol that's either a builtin or an imported
        value. This differs from a symbol defined in the block of code that happens
        to be in global scope, which will be given a static frame number.
    """

    name: str
    scope: ast_scope.scope.Scope

    @classmethod
    def parse(cls, x):
        """
        Parses a symbol.
        """
        if x.startswith("&"):
            name, scope = x[1:].split(":")
            return cls(name, scope)
        if x.startswith("g"):
            assert x.startswith("g_")
            return cls(x[2:], None)
        return None

    def render(self):
        if self.scope is None:
            return f"g_{self.name}"
        return f"&{self.name}:{self.scope}"


def create_descoper(code):
    """
    Creates a mapping from nodes to numerical ids for scopes.

    Args:
        code: The code.

    Returns:
        The descoper.
    """
    globs = true_globals(code)
    annot = ast_scope.annotate(code)
    scopes = []
    results = {}
    for node in ast.walk(code):
        if node in annot:
            if node in globs:
                results[node] = None
                continue
            if annot[node] not in scopes:
                scopes.append(annot[node])
            results[node] = scopes.index(annot[node])
    return results


def true_globals(node):
    """
    Get the true globals of a program.

    Args:
        node: The node.

    Returns:
        The true globals.
    """
    name = "_" + uuid.uuid4().hex
    wpd = wrap_ast(node, name)
    scope_info = ast_scope.annotate(wpd)
    return {
        x
        for x in scope_info
        if scope_info[x] == scope_info.global_scope
        if getattr(x, python_ast_tools.name_field(x)) != name
    }



from ..dsl.dsl_factory import DSLFactory


def basic_arith_dsl(lambdas=False):
    dslf = DSLFactory()
    dslf.concrete("+", "(i, i) -> i", lambda x, y: x + y)
    dslf.concrete("1", "() -> i", lambda: 1)
    if lambdas:
        dslf.lambdas()
    dslf.prune_to("i", "i -> i", "i -> i -> i", "(i, i) -> i", "(i, i) -> i -> i")
    return dslf.finalize()



import sys
import unittest

from parameterized import parameterized

import neurosym as ns

from .utils import cwq, fit_to, small_set_runnable_code_examples


class DefUseMaskTestGeneric(unittest.TestCase):
    def annotate_alternates(self, chosen, alts):
        self.assertIn(chosen, alts)
        mat = ns.python_def_use_mask.match_either_name_or_global(chosen)
        if not mat:
            return chosen
        name, scope = mat.group("name"), (
            mat.group("scope") if mat.group("typ") == "&" else "0"
        )
        # print(alts)
        alts = [ns.python_def_use_mask.match_either_name_or_global(alt) for alt in alts]
        # print([x for x in alts if x])
        alts = {x.group("name") for x in alts if x}
        alts.remove(name)
        alts = sorted(alts)
        if alts:
            name = f"{name}?{'$'.join(alts)}"
        return f"const-&{name}:{scope}~Name"

    def annotate_program(
        self,
        program,
        parser=ns.python_to_python_ast,
        convert_to_python=True,
    ):
        dfa, _, fam, _ = fit_to(
            [program], parser=parser, include_type_preorder_mask=False
        )
        annotated = ns.s_exp_to_python_ast(
            ns.render_s_expression(
                ns.annotate_with_alternate_symbols(
                    ns.to_type_annotated_ns_s_exp(parser(program), dfa, "M"),
                    fam.tree_distribution_skeleton,
                    self.annotate_alternates,
                )
            )
        )
        if convert_to_python:
            return annotated.to_python()
        return annotated


class DefUseMaskTest(DefUseMaskTestGeneric):
    def test_annotate_alternate_symbols(self):
        code = self.annotate_program("x = 2; y = x; z = y")
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                x?y$z = 2
                y?x$z = x
                z?x$y = y?x
                """
            ).strip(),
        )

    def test_duplicate_lhs(self):
        code = self.annotate_program("x, x = 2, 2")
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                x, x = 2, 2
                """
            ).strip(),
        )

    def test_subscript_on_lhs(self):
        code = self.annotate_program("x = [2, 3, 4]; x[2] = x[0]; y = 2")
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                x?y = [2, 3, 4]
                x[2] = x[0]
                y?x = 2
                """
            ).strip(),
        )

    def test_attribute_on_lhs(self):
        code = self.annotate_program("x = 2; y.z = 3; x = x")
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                x?y = 2
                y?x.z = 3
                x?y = x?y
                """
            ).strip(),
        )

    def test_tuple_list_on_lhs(self):
        code = self.annotate_program("[x, y] = 2, 3; x, y = x, y; z = x")
        print(code)
        past_310 = """
        [x?y$z, y?x$z] = (2, 3)
        x?y$z, y?x$z = (x?y, y?x)
        z?x$y = x?y
        """
        up_to_310 = """
        [x?y$z, y?x$z] = (2, 3)
        (x?y$z, y?x$z) = (x?y, y?x)
        z?x$y = x?y
        """
        self.assertEqual(
            code.strip(),
            cwq(up_to_310 if sys.version_info < (3, 11) else past_310).strip(),
        )

    def test_star_tuple_on_lhs(self):
        code = self.annotate_program("x, *y = [2, 3]; x = x")
        print(code)
        past_310 = """
        x?y, *y?x = [2, 3]
        x?y = x?y
        """
        up_to_310 = """
        (x?y, *y?x) = [2, 3]
        x?y = x?y
        """
        self.assertEqual(
            code.strip(),
            cwq(up_to_310 if sys.version_info < (3, 11) else past_310).strip(),
        )

    def test_basic_import(self):
        # the 2 in front is necessary to force the import to not be pulled
        code = self.annotate_program(
            cwq(
                """
                2
                import os
                import sys as y
                from collections import defaultdict
                from collections import defaultdict as z
                x = os
                x = os
                defaultdict, os, y, z
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                2
                import os?defaultdict
                import sys as y?z
                from collections import defaultdict?os
                from collections import defaultdict as z?y
                x?defaultdict$os$y$z = os?defaultdict$y$z
                x?defaultdict$os$y$z = os?defaultdict$x$y$z
                (defaultdict?os$x$y$z, os?defaultdict$x$y$z,
                    y?defaultdict$os$x$z, z?defaultdict$os$x$y)
                """
            ).strip(),
        )

    def test_function_call(self):
        code = self.annotate_program(
            cwq(
                """
                def f(x):
                    z = x
                    return x
                y = f(2)
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                def f?x$y$z(x?f$y$z):
                    z?f$x$y = x?f
                    return x?f$z
                y?f$x$z = f(2)
                """
            ).strip(),
        )

    def test_lambda(self):
        code = self.annotate_program(
            cwq(
                """
                x = 2
                lambda y, z=x: lambda a=y: x
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                x?a$y$z = 2
                lambda y?a$x$z, z?a$x$y=x: lambda a?x$y$z=y?x$z: x?a$y$z
                """
            ).strip(),
        )

    def test_function_call_arguments(self):
        code = self.annotate_program(
            cwq(
                """
                def f(w, /, x, *y, **z):
                    return x
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                def f?w$x$y$z(w?f$x$y$z, /, x?f$w$y$z, *y?f$w$x$z, **z?f$w$x$y):
                    return x?f$w$y$z
                """
            ).strip(),
        )

    def test_single_comprehension(self):
        code = self.annotate_program(
            cwq(
                """
                a = 2
                [b for b in range(a) if b == a]
                a = a
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                a?b$range = 2
                [b?a$range for b?a$range in range?a(a?range) if b?a$range == a?b$range]
                a?b$range = a?range
                """
            ).strip(),
        )

    def test_bunch_of_comprehensions(self):
        self.maxDiff = None
        code = self.annotate_program(
            cwq(
                """
                a = 2
                [b for b in range(a)]
                (c for c in range(a))
                {c for c in range(a)}
                {d: a for d in range(a)}
                [e + f + g for e in range(a) for f in range(e) for g in range(f)]
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                a?b$c$d$e$f$g$range = 2
                [b?a$range for b?a$c$d$e$f$g$range in range?a(a?range)]
                (c?a$range for c?a$b$d$e$f$g$range in range?a(a?range))
                {c?a$range for c?a$b$d$e$f$g$range in range?a(a?range)}
                {d?a$range: a?d$range for d?a$b$c$e$f$g$range in range?a(a?range)}
                [e?a$f$g$range + f?a$e$g$range + g?a$e$f$range
                    for e?a$b$c$d$f$g$range in range?a(a?range)
                    for f?a$b$c$d$e$g$range in range?a$e(e?a$range)
                    for g?a$b$c$d$e$f$range in range?a$e$f(f?a$e$range)]
                """
            ).strip(),
        )

    def test_for(self):
        self.maxDiff = None
        code = self.annotate_program(
            cwq(
                """
                x = [2]
                for y in x:
                    y
                z = x
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                x?y$z = [2]
                for y?x$z in x:
                    y?x
                z?x$y = x?y
                """
            ).strip(),
        )

    def test_import_at_top_level(self):
        # imports at top are global so not alternated
        code = self.annotate_program("import os; import sys as y; x = os; x = os; x, y")
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                import os
                import sys as y
                x?os$y = os?y
                x?os$y = os?x$y
                (x?os$y, y?os$x)
                """
            ).strip(),
        )

    def test_class(self):
        code = self.annotate_program(
            cwq(
                """
                class A:
                    x = A
                y = A
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                class A?x$y:
                    x?A$y = A
                y?A$x = A
                """
            ).strip(),
        )

    def test_import_inside_fn(self):
        code = self.annotate_program(
            cwq(
                """
                def f():
                    from collections import defaultdict
                    return defaultdict
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                def f?defaultdict():
                    from collections import defaultdict
                    return defaultdict?f
                """
            ).strip(),
        )

    def test_function_default(self):
        code = self.annotate_program(
            cwq(
                """
                y = 2
                z = 3
                def f(x=y):
                    return x
                z = z
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                y?f$x$z = 2
                z?f$x$y = 3

                def f?x$y$z(x?f$y$z=y?f$z):
                    return x?f$y$z
                z?f$x$y = z?f$y
                """
            ).strip(),
        )

    def test_exception_named(self):
        code = self.annotate_program(
            cwq(
                """
                try:
                    x = 2
                except Exception as e:
                    x = e
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                try:
                    x?Exception$e = 2
                except Exception?x as e:
                    x?Exception$e = e?Exception$x
                """
            ).strip(),
        )

    def test_exception_unnamed(self):
        code = self.annotate_program(
            cwq(
                """
                try:
                    x = 2
                except Exception:
                    x = x
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                try:
                    x?Exception = 2
                except Exception?x:
                    x?Exception = x?Exception
                """
            ).strip(),
        )

    def test_complicated_type_annot(self):
        code = self.annotate_program(
            cwq(
                """
                x: List[Dict[str, int]] = []
                """
            )
        )
        print(code)
        self.assertEqual(
            code.strip(),
            cwq(
                """
                x: List[Dict[str, int]] = []
                """
            ).strip(),
        )

    @parameterized.expand([(i,) for i in range(20)])
    def test_realistic(self, i):
        if i in {22, 31, 41, 57, 95, 100, 106, 109, 112, 114, 119, 181, 182}:
            # forward declaration of
            # input for 22/41/100/119
            # n for 31/57/112/114
            # m for 95
            # sp for 106
            # mp for 109
            # dp for 181
            # lo for 182 [this one's weird]
            return
        example = small_set_runnable_code_examples()[i]["solution"]
        print(example)
        code = self.annotate_program(example)
        print(code)



import ast

from neurosym.python_dsl.convert_python.python_ast import (
    LeafAST,
    ListAST,
    NodeAST,
    PythonAST,
)
from neurosym.python_dsl.convert_python.symbol import PythonSymbol


def make_constant(leaf):
    """
    Create a constant PythonAST from the given leaf value (which must be a python constant).
    """
    assert not isinstance(leaf, PythonAST), leaf
    return NodeAST(typ=ast.Constant, children=[LeafAST(leaf=leaf), LeafAST(leaf=None)])


def make_name(name_node):
    """
    Create a name PythonAST from the given name node containing a symbol.
    """
    assert isinstance(name_node, LeafAST) and isinstance(
        name_node.leaf, PythonSymbol
    ), name_node
    return NodeAST(
        typ=ast.Name,
        children=[
            name_node,
            NodeAST(typ=ast.Load, children=[]),
        ],
    )


def make_call(name_sym, *arguments):
    """
    Create a call PythonAST from the given symbol and arguments.

    In this case, the symbol must be a symbol representing a name.
    """
    assert isinstance(name_sym, PythonSymbol), name_sym
    return NodeAST(
        typ=ast.Call,
        children=[
            make_name(LeafAST(name_sym)),
            ListAST(children=arguments),
            ListAST(children=[]),
        ],
    )


def make_expr_stmt(expr):
    """
    Create an expression statement PythonAST from the given expression.
    """
    return NodeAST(typ=ast.Expr, children=[expr])



from dataclasses import dataclass, field
from typing import Dict

from neurosym.programs.s_expression import SExpression


@dataclass(unsafe_hash=True)
class DSLSearchNode:
    program: SExpression
    # metadata includes things related to the search graph, e.g., depth of the node
    metadata: Dict[str, object] = field(hash=False)



import copy
from dataclasses import dataclass
from typing import Callable, Dict, List, Tuple

from neurosym.dsl.dsl import DSL
from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.preorder_mask.undos import chain_undos
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution

from .extra_var import ExtraVar, canonicalized_python_name_leaf_regex
from .handler import DefaultHandler, HandlerPuller, default_handler
from .names import NAME_REGEX
from .special_case_symbol_predicate import NameEPredicate, SpecialCaseSymbolPredicate


@dataclass
class DefUseMaskConfiguration:
    """
    Configuration for the DefUseMask.
    """

    dfa: Dict
    node_hooks: Dict[str, HandlerPuller]

    def get_hook(self, symbol):
        prefixes = [x for x in self.node_hooks if symbol.startswith(x)]
        if not prefixes:
            return None
        assert len(prefixes) == 1, f"Multiple hooks found for {symbol}: {prefixes}"
        return self.node_hooks[prefixes[0]]

    def pull_handler(
        self,
        position: int,
        symbol: str,
        mask: "DefUseChainPreorderMask",
        defined_production_idxs: List[int],
        handler_fn=default_handler,
    ):
        hook = self.get_hook(symbol)
        if hook is None:
            return None
        return hook.pull_handler(
            position,
            symbol,
            mask,
            defined_production_idxs,
            self,
            handler_fn=handler_fn,
        )


class DefUseChainPreorderMask(PreorderMask):
    """
    Preorder mask that filters out symbols that are not defined at a given position.

    The idea is to have a stack of Handler objects, one for each node in the syntax tree.

    Args:
        tree_dist: The tree distribution that the mask is applied to.
        dsl: The domain-specific language that the mask is applied to.
        config: The configuration for the mask.
        special_case_predicate_fns: A list of functions that return special case predicates.
    """

    def __init__(
        self,
        tree_dist: TreeDistribution,
        dsl: DSL,
        config: DefUseMaskConfiguration,
        special_case_predicate_fns: Tuple[
            Callable[[TreeDistribution], SpecialCaseSymbolPredicate]
        ] = (),
    ):
        super().__init__(tree_dist)
        self.dsl = dsl
        self.idx_to_name = []
        for x, _ in self.tree_dist.symbols:
            mat = NAME_REGEX.match(x)
            self.idx_to_name.append(mat.group("name") if mat else None)

        self.special_case_predicates = [
            fn(self.tree_dist) for fn in (NameEPredicate, *special_case_predicate_fns)
        ]

        self.handlers = []
        self.config = config

    def currently_defined_indices(self):
        """
        Return the indices of the symbols that are currently defined.
        """
        return self.handlers[-1].currently_defined_indices()

    def compute_mask(self, position: int, symbols: List[int]) -> List[bool]:
        """
        Compute the mask for the given position and symbols. If the last handler is
            defining, then all symbols are valid. Otherwise, only the symbols that
            match the handler's names are valid.
        """
        return self.handlers[-1].compute_mask(
            position, symbols, self.idx_to_name, self.special_case_predicates
        )

    def on_entry(self, position: int, symbol: int) -> Callable[[], None]:
        """
        Updates the stack of handlers when entering a node.
        """
        if not self.handlers:
            assert position == symbol == 0
            self.handlers.append(DefaultHandler(self, [], self.config))
            return self.handlers.pop
        new_handler, undo = self.handlers[-1].on_child_enter(position, symbol)
        self.handlers.append(new_handler)

        return chain_undos(
            [
                getattr(new_handler, "__undo__init__", lambda: None),
                undo,
                self.handlers.pop,
            ]
        )

    def on_exit(self, position: int, symbol: int) -> Callable[[], None]:
        """
        Updates the stack of handlers when exiting a node.
        """
        popped = self.handlers.pop()
        undo_pop = lambda: self.handlers.append(popped)
        if not self.handlers:
            assert position == symbol == 0
            return undo_pop
        undo = self.handlers[-1].on_child_exit(position, symbol, popped)

        return chain_undos([undo_pop, undo])

    def with_handler(self, handler_fn):
        """
        With a new handler function, return a new mask.
        """
        mask_copy = copy.copy(self)
        handler = handler_fn(mask_copy)
        mask_copy.handlers = [handler]
        return mask_copy

    def id_to_name_and_arity(self, symbol_id):
        """
        Convert the symbol ID to a string of the symbol name, and the arity of the symbol.
        """

        if isinstance(symbol_id, ExtraVar):
            return symbol_id.leaf_name(), 0
        return self.tree_dist.symbols[symbol_id]

    def id_to_name(self, symbol_id):
        """
        Convert the symbol ID to a string.
        """
        return self.id_to_name_and_arity(symbol_id)[0]

    def name_to_id(self, name: str):
        """
        Convert the string to a symbol ID.
        """
        if canonicalized_python_name_leaf_regex.match(name):
            assert name not in self.tree_dist.symbol_to_index
            evar = ExtraVar.from_name(name)
            if evar is not None:
                return evar
        return self.tree_dist.symbol_to_index[name]



from abc import ABC, abstractmethod
from typing import List, TypeVar

import numpy as np
import torch

from neurosym.dsl.dsl import DSL
from neurosym.program_dist.enumeration_chunk_size import DEFAULT_CHUNK_SIZE
from neurosym.programs.s_expression import SExpression

ProgramDistribution = TypeVar("ProgramDistribution")
ProgramDistributionBatch = TypeVar("ProgramDistributionBatch")
ProgramCountsTensorBatch = TypeVar("ProgramCountsTensorBatch")


class ProgramDistributionFamily(ABC):
    """
    Abstract class representing a family of distributions over programs.
    """

    @abstractmethod
    def underlying_dsl(self) -> DSL:
        """
        Returns the DSL that this distribution is over.
        """

    @abstractmethod
    def parameters_shape(self) -> List[int]:
        """
        Returns the shape of the parameters of this distribution.
        """

    @abstractmethod
    def with_parameters(self, parameters: torch.Tensor) -> ProgramDistributionBatch:
        """
        Initializes a distribution from the given parameters. The parameters
            should have the shape (batch_size, *self.parameters_shape()).
        """

    @abstractmethod
    def count_programs(self, data: List[List[SExpression]]) -> ProgramCountsTensorBatch:
        """
        For each program, count its components' occurrences in the data. This
            depends on the type of distribution.
        """

    @abstractmethod
    def counts_to_distribution(
        self, counts: ProgramCountsTensorBatch
    ) -> ProgramDistributionBatch:
        """
        Converts the counts to a distribution.
        """

    @abstractmethod
    def parameter_difference_loss(
        self, parameters: torch.tensor, actual: ProgramCountsTensorBatch
    ) -> torch.float32:
        """
        Returns the loss between the parameters and actual counts, for
            several count tensors. Keep in mind that the parameters
            here should have a batch dimension, and
            parameters.shape[0] == len(actual).
        """

    @abstractmethod
    def sample(
        self,
        dist: ProgramDistribution,
        rng: np.random.RandomState,
        *,
        depth_limit=float("inf"),
    ) -> SExpression:
        """
        Samples programs from this distribution.
        """

    @abstractmethod
    def enumerate(
        self,
        dist: ProgramDistribution,
        *,
        min_likelihood: float = float("-inf"),
        chunk_size: float = DEFAULT_CHUNK_SIZE,
    ):
        """
        Enumerate all programs using iterative deepening. Yields (program, likelihood).

        Args:
            dist: The distribution to sample from.
            chunk_size: The amount of likelihood to consider at once. If this is
                too small, we will spend a lot of time doing the same work over and
                over again. If this is too large, we will spend a lot of time
                doing work that we don't need to do.
        """



def defining_statement_handlers():
    """
    Map from statement name to handler class for defining statements.
    """
    # pylint: disable=cyclic-import
    from .chained_definitions_handler import chained_definition_handlers
    from .defining_construct_handler import defining_construct_handler
    from .defining_statement_handler import (
        AnnAssignHandler,
        AssignHandler,
        ForHandler,
        ImportFromHandler,
        ImportHandler,
        LambdaHandler,
    )
    from .except_handler_handler import ExceptHandlerHandler

    return {
        x.name: x
        for x in [
            AssignHandler,
            AnnAssignHandler,
            ImportHandler,
            ImportFromHandler,
            ForHandler,
            LambdaHandler,
            ExceptHandlerHandler,
            *defining_construct_handler,
            *chained_definition_handlers,
        ]
    }



import ast
from types import NoneType
from typing import List, Union

from increase_recursionlimit import increase_recursionlimit

from neurosym.python_dsl import python_ast_tools

from .python_ast import (
    LeafAST,
    ListAST,
    NodeAST,
    PythonAST,
    SequenceAST,
    SliceElementAST,
    StarrableElementAST,
)
from .symbol import PythonSymbol, create_descoper


def python_body_to_parsed_ast(x: List[ast.AST], descoper: dict) -> SequenceAST:
    """
    Convert a python body to a PythonAST object.

    Args:
        x: The python body.
        descoper: The descoper to use.

    Returns:
        The parsed PythonAST.
    """
    assert isinstance(x, list), str(x)
    x = [python_ast_to_parsed_ast(x, descoper) for x in x]
    return SequenceAST("/seq", x)


def python_ast_to_parsed_ast(x, descoper: dict) -> PythonAST:
    """
    Convert an ast.AST object to a PythonAST object.

    Args:
        x: The ast.AST object.
        descoper: The descoper to use.

    Returns:
        The parsed PythonAST.
    """
    if isinstance(x, ast.AST):
        result = []
        for f in python_ast_tools.fields_for_node(x):
            el = getattr(x, f)
            if x in descoper and f == python_ast_tools.name_field(x):
                assert isinstance(el, str), (x, f, el)
                result.append(LeafAST(PythonSymbol(el, descoper[x])))
            else:
                if f == "slice":
                    result.append(
                        SliceElementAST(python_ast_to_parsed_ast(el, descoper))
                    )
                elif python_ast_tools.field_is_starrable(type(x), f):
                    out = python_ast_to_parsed_ast(el, descoper)
                    out = ListAST([StarrableElementAST(x) for x in out.children])
                    result.append(out)
                elif python_ast_tools.field_is_body(type(x), f):
                    result.append(python_body_to_parsed_ast(el, descoper))
                else:
                    result.append(python_ast_to_parsed_ast(el, descoper))
        return NodeAST(type(x), result)
    if isinstance(x, list):
        return ListAST([python_ast_to_parsed_ast(x, descoper) for x in x])
    if x is None or x is Ellipsis or isinstance(x, (int, float, complex, str, bytes)):
        return LeafAST(x)
    raise ValueError(f"Unsupported node {x}")


def python_to_python_ast(
    code: Union[str, ast.AST], descoper: Union[NoneType, dict] = None
) -> PythonAST:
    """
    Parse the given python code into a PythonAST. If the code is a string,
        it is first parsed into an AST.

    Args:
        code: The python code to parse.
        descoper: The descoper to use. If None, a new one is created.

    Returns:
        The parsed PythonAST.
    """

    with increase_recursionlimit():
        if isinstance(code, str):
            code = ast.parse(code)
        code = python_ast_to_parsed_ast(
            code,
            descoper if descoper is not None else create_descoper(code),
        )
        return code



import unittest

import neurosym as ns


class TestUnify(unittest.TestCase):
    def test_unify_atomic(self):
        self.assertEqual(
            ns.parse_type("int").unify(ns.parse_type("int")),
            dict(),
        )

    def test_unify_atomic_fail(self):
        with self.assertRaises(ns.UnificationError):
            ns.parse_type("int").unify(ns.parse_type("float"))

    def test_unify_var_atomic(self):
        self.assertEqual(
            ns.parse_type("#a").unify(ns.parse_type("int")),
            dict(a=ns.parse_type("int")),
        )

    def test_unify_tensor(self):
        self.assertEqual(
            ns.parse_type("{int, 3, 4}").unify(ns.parse_type("{int, 3, 4}")),
            dict(),
        )

    def test_unify_tensor_var(self):
        self.assertEqual(
            ns.parse_type("{#a, 3, 4}").unify(ns.parse_type("{int, 3, 4}")),
            dict(a=ns.parse_type("int")),
        )
        self.assertEqual(
            ns.parse_type("#a").unify(ns.parse_type("{int, 3, 4}")),
            dict(a=ns.parse_type("{int, 3, 4}")),
        )

    def test_unify_tensor_fail(self):
        with self.assertRaises(ns.UnificationError):
            ns.parse_type("{int, 3, 4}").unify(ns.parse_type("{int, 3, 5}"))

        with self.assertRaises(ns.UnificationError):
            ns.parse_type("{int, 3, 4}").unify(ns.parse_type("{float, 3, 4}"))

    def test_unify_list(self):
        self.assertEqual(
            ns.parse_type("[int]").unify(ns.parse_type("[int]")),
            dict(),
        )

    def test_unify_list_var(self):
        self.assertEqual(
            ns.parse_type("#a").unify(ns.parse_type("[int]")),
            dict(a=ns.parse_type("[int]")),
        )
        self.assertEqual(
            ns.parse_type("[#a]").unify(ns.parse_type("[int]")),
            dict(a=ns.parse_type("int")),
        )
        self.assertEqual(
            ns.parse_type("[#a]").unify(ns.parse_type("[[int]]")),
            dict(a=ns.parse_type("[int]")),
        )

    def test_unify_list_fail(self):
        with self.assertRaises(ns.UnificationError):
            ns.parse_type("[int]").unify(ns.parse_type("[float]"))

    def test_unify_arrow(self):
        self.assertEqual(
            ns.parse_type("(int, float) -> bool").unify(
                ns.parse_type("(int, float) -> bool")
            ),
            dict(),
        )

    def test_unify_arrow_var(self):
        self.assertEqual(
            ns.parse_type("#a").unify(ns.parse_type("(int, float) -> bool")),
            dict(a=ns.parse_type("(int, float) -> bool")),
        )
        self.assertEqual(
            ns.parse_type("(#a, float) -> bool").unify(
                ns.parse_type("(int, float) -> bool")
            ),
            dict(a=ns.parse_type("int")),
        )
        self.assertEqual(
            ns.parse_type("(#a, #a) -> bool").unify(
                ns.parse_type("(int, int) -> bool")
            ),
            dict(a=ns.parse_type("int")),
        )

    def test_unify_arrow_fail(self):
        with self.assertRaises(ns.UnificationError):
            ns.parse_type("(int, float) -> bool").unify(
                ns.parse_type("(int, int) -> bool")
            )

        with self.assertRaises(ns.UnificationError):
            ns.parse_type("(int, float) -> bool").unify(
                ns.parse_type("(float, int) -> bool")
            )

        with self.assertRaises(ns.UnificationError):
            ns.parse_type("(int, float) -> bool").unify(
                ns.parse_type("(#a, #a) -> bool")
            )

        with self.assertRaises(ns.UnificationError):
            ns.parse_type("int -> bool").unify(ns.parse_type("#a -> #a"))



import json
import unittest

import numpy as np

from .utils import execute_notebook


class TestTutorial1(unittest.TestCase):
    def test_tutorial_1_classification(self):
        result = execute_notebook(
            "tutorial/tutorial1_classification.ipynb",
            suffix="import json; print('*' * 80); print(json.dumps(lin.weight.tolist()))",
        )

        *_, stars, weight, _ = result.split("\n")
        assert stars == "*" * 80
        weight = np.array(json.loads(weight))
        x, y = weight[1] - weight[0]
        ang = np.arctan2(y, x)
        ang = np.degrees(ang)

        # The angle should be between -80 and -100 degrees
        # representing an downward normal vector
        self.assertGreaterEqual(ang, -100)
        self.assertLessEqual(ang, -80)

    def test_tutorial_1_regression(self):
        """
        Counts the number of modules in the solution. Should be more than 1 as a linear
        layer won't be able to approximate a non-linear function
        """
        result = execute_notebook(
            "tutorial/tutorial1_regression.ipynb",
            suffix="import json; print('*' * 80); "
            + "print(json.dumps([len(module.contained_modules)]))",
        )

        *_, stars, len_module, _ = result.split("\n")
        assert stars == "*" * 80
        n_layers = json.loads(len_module)[0]
        self.assertGreaterEqual(n_layers, 1)



from typing import Callable, Tuple

from ..handler import ConstructHandler, Handler
from .defining_statement_handler import DefiningStatementHandler


class ComprehensionExpressionHandler(ConstructHandler):
    """
    Handles comprehensions. This isn't a DefiningStatementHandler because
        each comprehensions' defined symbols can be used in the next
        comprehension.
    """

    # keep in mind that the ordering means generators come first

    def __init__(self, mask, defined_production_idxs, config):
        # copy the valid symbols so changes don't affect the parent
        super().__init__(mask, defined_production_idxs[:], config)
        self.defined_symbols = []

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        if position == self.child_fields["generators"]:
            return (
                GeneratorsHandler(self.mask, self.defined_production_idxs, self.config),
                lambda: None,
            )
        return super().on_child_enter(position, symbol)

    def is_defining(self, position: int) -> bool:
        return False


class ListComprehensionHandler(ComprehensionExpressionHandler):
    name = "ListComp~E"


class SetComprehensionHandler(ListComprehensionHandler):
    name = "SetComp~E"


class GeneratorExprHandler(ListComprehensionHandler):
    name = "GeneratorExp~E"


class DictComprehensionHandler(ComprehensionExpressionHandler):
    name = "DictComp~E"


class GeneratorsHandler(Handler):
    """
    Handles a list of generators, each treated as a defining statement.
    """

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        return (
            ComprehensionHandler(self.mask, self.defined_production_idxs, self.config),
            lambda: None,
        )

    def is_defining(self, position: int) -> bool:
        return False


class ComprehensionHandler(DefiningStatementHandler):
    name = "comprehension~C"
    targeted = ["target"]
    define_symbols_on_exit = "iter"


chained_definition_handlers = [
    ListComprehensionHandler,
    SetComprehensionHandler,
    GeneratorExprHandler,
    DictComprehensionHandler,
]



import unittest

import neurosym as ns

from .bigram_test import fam_with_ordering, fam_with_ordering_231, fam_with_vars


class TreeDistributionTest(unittest.TestCase):
    def collect_symbols(self, family, program):
        return ns.render_s_expression(
            ns.annotate_with_alternate_symbols(
                ns.parse_s_expression(program), family.tree_distribution_skeleton
            )
        )

    def test_basic_program(self):
        self.assertEqual(
            self.collect_symbols(fam_with_vars, "(+ (1) (1))"),
            "(+/+,1,2,call (1/+,1,2,call) (1/+,1,2,call))",
        )

    def test_program_with_variables(self):
        self.assertEqual(
            self.collect_symbols(fam_with_vars, "(call (lam ($0_0)) (1))"),
            "(call/+,1,2,call (lam/lam ($0_0/$0_0,+,1,2,call)) (1/+,1,2,call))",
        )

    def test_program_with_ordering(self):
        self.assertEqual(
            self.collect_symbols(fam_with_ordering, "(+ (1) (2) (3))"),
            "(+/+ (1/1) (2/2) (3/3))",
        )

    def test_program_with_ordering_231(self):
        self.assertEqual(
            self.collect_symbols(fam_with_ordering_231, "(+ (2) (3) (1))"),
            "(+/+ (2/2) (3/3) (1/1))",
        )



"""
We want to enumerate programs as trees. Specifically, we have some distribution
    P(tree) = prod_{node in tree} P(node | ancestors(node)[:limit]).

Here, limit is a value that the distribution guarantees us. This is useful
    because we can use this to optimize our enumeration algorithm (since it
    reduces the amount of memory we need to save for each hole).

Our algorithm here is based on iterative deepening. We have a method that
    keeps a minimum and maximum likelihood, and enumerates all programs
    that are within this likelihood range.

Likelihood is defined as the log probability of the program.
"""

import itertools
from typing import List, Tuple

import numpy as np

from neurosym.program_dist.enumeration_chunk_size import DEFAULT_CHUNK_SIZE
from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression


def enumerate_tree_dist(
    tree_dist: TreeDistribution,
    *,
    chunk_size: float = DEFAULT_CHUNK_SIZE,
    min_likelihood: float = float("-inf"),
):
    """
    Enumerate all programs using iterative deepening.

    Args:
        tree_dist: The distribution to sample from.
        chunk_size: The amount of likelihood to consider at once. If this is
            too small, we will spend a lot of time doing the same work over and
            over again. If this is too large, we will spend a lot of time
            doing work that we don't need to do.
    """
    for chunk in itertools.count(1):
        likelihood_bound = -chunk * chunk_size
        preorder_mask = tree_dist.mask_constructor(tree_dist)
        preorder_mask.on_entry(0, 0)
        for program, likelihood in enumerate_tree_dist_dfs(
            tree_dist, likelihood_bound, ((0, 0),), preorder_mask
        ):
            if (
                max(likelihood_bound, min_likelihood)
                < likelihood
                <= likelihood_bound + chunk_size
            ):
                yield program, likelihood
        if likelihood_bound <= min_likelihood:
            return


def enumerate_tree_dist_dfs(
    tree_dist: TreeDistribution,
    min_likelihood: float,
    parents: Tuple[Tuple[int, int], ...],
    preorder_mask: PreorderMask,
):
    """
    Enumerate all programs that are within the likelihood range, with the given parents.
    """

    if min_likelihood > 0:
        # We can stop searching deeper.
        return

    assert len(parents) <= tree_dist.limit
    assert isinstance(
        preorder_mask, PreorderMask
    ), f"{preorder_mask} is not a PreorderMask"

    # Performed recursively for now.
    syms, log_probs = tree_dist.likelihood_arrays[parents]
    position = parents[-1][1]
    mask = preorder_mask.compute_mask(position, syms)
    denominator = np.logaddexp.reduce(log_probs[mask])
    for node, likelihood in zip(syms[mask], log_probs[mask] - denominator):
        new_parents = parents + (node,)
        new_parents = new_parents[-tree_dist.limit :]
        symbol, arity = tree_dist.symbols[node]
        undo_entry = preorder_mask.on_entry(position, node)
        for children, child_likelihood in enumerate_children_and_likelihoods_dfs(
            tree_dist,
            min_likelihood - likelihood,
            parents,
            node,
            num_children=arity,
            starting_index=0,
            order=tree_dist.ordering.order(node, arity),
            preorder_mask=preorder_mask,
        ):
            undo_exit = preorder_mask.on_exit(position, node)
            yield SExpression(
                symbol, [children[i] for i in range(arity)]
            ), child_likelihood + likelihood
            undo_exit()
        undo_entry()


def enumerate_children_and_likelihoods_dfs(
    tree_dist: TreeDistribution,
    min_likelihood: float,
    parents: Tuple[Tuple[int, int], ...],
    most_recent_parent: int,
    num_children: int,
    starting_index: int,
    order: List[int],
    preorder_mask: PreorderMask,
):
    """
    Enumerate all children and their likelihoods.
    """

    if starting_index == num_children:
        yield {}, 0
        return
    new_parents = parents + ((most_recent_parent, order[starting_index]),)
    new_parents = new_parents[-tree_dist.limit :]

    for first_child, first_likelihood in enumerate_tree_dist_dfs(
        tree_dist, min_likelihood, new_parents, preorder_mask
    ):
        for (
            rest_children,
            rest_likelihood,
        ) in enumerate_children_and_likelihoods_dfs(
            tree_dist,
            min_likelihood - first_likelihood,
            parents,
            most_recent_parent,
            num_children,
            starting_index + 1,
            order,
            preorder_mask,
        ):
            yield {
                order[starting_index]: first_child,
                **rest_children,
            }, first_likelihood + rest_likelihood



from abc import ABC, abstractmethod
from types import NoneType
from typing import Dict, Iterator, List, Union

from neurosym.programs.s_expression import SExpression


class NodeOrdering(ABC):
    """
    Represents a technique for ordering the subnodes of a node in a tree.
    """

    def __init__(self, dist):
        self.dist = dist

    @abstractmethod
    def compute_order(self, root_sym_idx: int) -> Union[List[int], NoneType]:
        """
        Orders the subnodes of the node with the given symbol index.

        None is equivalent to list(range(n_subnodes)).
        """

    def order(self, root_sym_idx: int, n_subnodes: int) -> List[int]:
        """
        Orders the subnodes of the node with the given symbol index.
        """
        assert not isinstance(
            root_sym_idx, str
        ), f"Expected int, got {root_sym_idx} of type {type(root_sym_idx)}"
        order = self.compute_order(root_sym_idx)
        if order is None:
            return range(n_subnodes)
        return order

    def traverse_preorder(self, node: SExpression) -> Iterator[SExpression]:
        """
        Orders the subnodes of the given node.
        """
        yield node
        sym = self.dist.symbol_to_index[node.symbol]
        for i in self.order(sym, len(node.children)):
            yield from self.traverse_preorder(node.children[i])


class DictionaryNodeOrdering(NodeOrdering):
    """
    Orders the subnodes of a node according to a dictionary.
    """

    def __init__(self, dist, ordering: Dict[str, List[int]], tolerate_missing=False):
        super().__init__(dist)
        if not tolerate_missing:
            assert set(ordering.keys()).issubset(dist.symbol_to_index.keys())
        self.ordering = {
            dist.symbol_to_index[k]: vs
            for k, vs in ordering.items()
            if k in dist.symbol_to_index
        }

    def compute_order(self, root_sym_idx: int) -> Union[List[int], NoneType]:
        return self.ordering.get(root_sym_idx, None)


class DefaultNodeOrdering(NodeOrdering):
    """
    Orders the subnodes of a node according to their symbol indices.
    """

    def compute_order(self, root_sym_idx: int) -> Union[List[int], NoneType]:
        return None



from typing import Callable, Tuple

from neurosym.program_dist.tree_distribution.preorder_mask.undos import (
    chain_undos,
    remove_last_n_elements,
)

from ..handler import ConstructHandler, Handler


class DefiningStatementHandler(ConstructHandler):
    """
    Represents a statement that defines symbols.
    """

    # these fields must be defined in the subclass
    targeted: list[str] = None
    # the field after which the symbols are defined
    define_symbols_on_exit: str = None

    def __init__(self, mask, defined_production_idxs, config):
        super().__init__(mask, defined_production_idxs, config)
        assert isinstance(self.name, str)
        assert isinstance(self.targeted, list)
        assert isinstance(self.define_symbols_on_exit, str)
        self._targeted_positions = [self.child_fields[child] for child in self.targeted]
        self.defined_symbols = []

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        if position in self._targeted_positions:
            return self.target_child(position, symbol)
        return super().on_child_enter(position, symbol)

    def on_child_exit(
        self, position: int, symbol: int, child: Handler
    ) -> Callable[[], None]:
        undos = []
        if position in self._targeted_positions:
            if len(set(child.defined_symbols)) != len(child.defined_symbols):
                raise ValueError(
                    f"Duplicate symbols defined in child {child}: {child.defined_symbols}"
                )
            self.defined_symbols += child.defined_symbols
            undos.append(
                remove_last_n_elements(self.defined_symbols, len(child.defined_symbols))
            )
        if position == self.child_fields[self.define_symbols_on_exit]:
            current = set(self.defined_production_idxs)
            additional_symbols = [x for x in self.defined_symbols if x not in current]
            self.defined_production_idxs += additional_symbols
            undos.append(
                remove_last_n_elements(
                    self.defined_production_idxs, len(additional_symbols)
                )
            )
        undos.append(super().on_child_exit(position, symbol, child))
        return chain_undos(undos)

    def is_defining(self, position: int) -> bool:
        return position in self._targeted_positions


class ChildFrameCreatorHandler(DefiningStatementHandler):
    def __init__(self, mask, defined_production_idxs, config):
        self.original_defined_production_idxs = defined_production_idxs
        super().__init__(mask, list(defined_production_idxs), config)


class AssignHandler(DefiningStatementHandler):
    name = "Assign~S"
    targeted = ["targets"]
    define_symbols_on_exit = "type_comment"


class AnnAssignHandler(DefiningStatementHandler):
    name = "AnnAssign~S"
    targeted = ["target"]
    define_symbols_on_exit = "simple"


class ForHandler(DefiningStatementHandler):
    name = "For~S"
    targeted = ["target"]
    define_symbols_on_exit = "iter"


class ImportHandler(DefiningStatementHandler):
    name = "Import~S"
    targeted = ["names"]
    define_symbols_on_exit = "names"


class ImportFromHandler(ImportHandler):
    name = "ImportFrom~S"


class LambdaHandler(ChildFrameCreatorHandler):
    name = "Lambda~E"
    targeted = ["args"]
    define_symbols_on_exit = "args"



from abc import ABC, abstractmethod
from typing import List

from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution

from .names import GLOBAL_REGEX


class SpecialCaseSymbolPredicate(ABC):
    """
    Handles a special case symbol matching predicate, i.e., a predicate that applies only in
        special cases.
    """

    def __init__(self, tree_dist: TreeDistribution):
        self.tree_dist = tree_dist

    @abstractmethod
    def applies(self, symbol: int) -> bool:
        """
        Whether or not the predicate applies to the symbol.
        """

    @abstractmethod
    def compute(self, symbol: int, names: List[str]) -> bool:
        """
        Compute the mask for the given symbol.
        """


class NameEPredicate(SpecialCaseSymbolPredicate):
    """
    Predicate that applies to the name_e symbol.
    """

    def __init__(self, tree_dist: TreeDistribution):
        super().__init__(tree_dist)
        self.has_global_available = any(
            GLOBAL_REGEX.match(x) for x, _ in tree_dist.symbols
        )
        self.name_e = tree_dist.symbol_to_index.get("Name~E", -1)

    def applies(self, symbol: int) -> bool:
        return symbol == self.name_e

    def compute(self, symbol: int, names: List[str]) -> bool:
        return self.has_global_available or len(names) > 0



from typing import Callable, Tuple

from neurosym.program_dist.tree_distribution.preorder_mask.undos import chain_undos

from ..handler import ConstructHandler, Handler


class ExceptHandlerHandler(ConstructHandler):
    """
    Handles an exception statement, which defines a name in the except block.
    """

    name = "ExceptHandler~EH"

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        undos = []
        if self.is_defining(position):
            if self.mask.id_to_name(symbol) != "const-None~NullableName":
                self.defined_production_idxs.append(symbol)
                undos.append(self.defined_production_idxs.pop)
        handler, undo = super().on_child_enter(position, symbol)
        undos.append(undo)
        return handler, chain_undos(undos)

    def is_defining(self, position: int) -> bool:
        return position == self.child_fields["name"]



"""
Test dsl/differentiable_dsl.py with NEAR search graph.

We conduct the following tests:
- Sanity check: We can find a simple program.
- BFS: We can find a simple program with BFS.
- Astar: We can find a simple program with bounded Astar search.
- Enumerate: We can enumerate all programs of a certain size.
NEAR Integration tests.
"""

import unittest

import torch

import neurosym as ns
from neurosym.examples import near

from .utils import assertDSLEnumerable


class TestNEARSimpleDSL(unittest.TestCase):
    def test_simple_dsl_bfs(self):
        """
        input: x: tensor[input_dim]
        goal: f(x) = 4
        expected node.program:
            (int_int_add (one) (int_int_add (one) (int_int_add (one) (one))))
            ie. 1 + (1 + (1 + 1))
        """
        self.maxDiff = None
        input_dim = 10
        dsl = near.differentiable_arith_dsl(input_dim)
        g = near.near_graph(
            dsl,
            ns.parse_type("f"),
            is_goal=lambda x: dsl.compute(dsl.initialize(x.program)) == 4,
        )
        node = next(ns.search.bfs(g)).program
        self.assertEqual(
            str(node),
            str(
                ns.SExpression(
                    symbol="int_int_add",
                    children=(
                        ns.SExpression(symbol="one", children=()),
                        ns.SExpression(
                            symbol="int_int_add",
                            children=(
                                ns.SExpression(symbol="one", children=()),
                                ns.SExpression(
                                    symbol="int_int_add",
                                    children=(
                                        ns.SExpression(symbol="one", children=()),
                                        ns.SExpression(symbol="one", children=()),
                                    ),
                                ),
                            ),
                        ),
                    ),
                )
            ),
        )

    def test_simple_dsl_astar(self):
        """
        input: x: tensor[input_dim]
        goal: f(x) = [4] * input_dim
        expected node.program:
            (ones input_dim)
        """
        self.maxDiff = None
        input_size = 10
        dsl = near.differentiable_arith_dsl(input_size)
        fours = torch.full((input_size,), 4.0)

        def checker(x):
            x = x.program
            xx = dsl.compute(dsl.initialize(x))
            if isinstance(xx, torch.Tensor):
                return torch.all(torch.eq(xx, fours))
            return False

        g = near.near_graph(dsl, ns.parse_type("{f, 10}"), is_goal=checker)

        def cost(x):
            if isinstance(x.program, ns.SExpression) and x.program.children:
                return len(str(x.program.children[0]))
            return 0

        node = next(ns.search.bounded_astar(g, cost, max_depth=7)).program
        self.assertEqual(node.children[0], ns.SExpression(symbol="ones", children=()))

    def test_simple_dsl_enumerate(self):
        """
        Enumerate all programs in dsl upto fixed depth. This test case makes
        sure all DSL combinations upto a fixed depth are valid.
        """
        self.maxDiff = None
        dsl = near.differentiable_arith_dsl(10)

        assertDSLEnumerable(dsl, "$fL")



def chain_undos(undos):
    """
    Run a series of undo functions. The undos are run in reverse order,
        so the last undo function is run first.
    """

    def undo():
        for undo in undos[::-1]:
            undo()

    return undo


def remove_last_n_elements(lst, n):
    """
    Remove the last n elements from a list.
    """

    def run():
        if n > 0:
            del lst[-n:]

    return run



from .extra_var import ExtraVar, canonicalized_python_name_as_leaf
from .handler import Handler, HandlerPuller, default_handler
from .mask import DefUseChainPreorderMask, DefUseMaskConfiguration
from .names import GLOBAL_REGEX, NAME_REGEX, match_either_name_or_global
from .ordering import PythonNodeOrdering, python_ordering_dictionary
from .special_case_symbol_predicate import SpecialCaseSymbolPredicate
from .target_handler import TargetHandler



import unittest

import neurosym as ns


class TestReplaceSymbol(unittest.TestCase):
    def test_basic(self):
        s_exp = ns.parse_s_expression("(foo (bar) (baz))")
        id_to_symbol = {id(s_exp.children[0]): "abc", id(s_exp): "def"}
        replaced = s_exp.replace_symbols_by_id(id_to_symbol)
        self.assertEqual(ns.render_s_expression(replaced), "(def (abc) (baz))")

    def test_with_duplicates(self):
        s_exp = ns.parse_s_expression("(foo (bar) (bar) (bar))")
        id_to_symbol = {id(s_exp.children[0]): "abc", id(s_exp.children[1]): "def"}
        replaced = s_exp.replace_symbols_by_id(id_to_symbol)
        self.assertEqual(ns.render_s_expression(replaced), "(foo (abc) (def) (bar))")


class TestReplaceNode(unittest.TestCase):
    def test_basic(self):
        s_exp = ns.parse_s_expression("(foo (bar) (baz))")
        id_to_node = {
            id(s_exp.children[0]): ns.parse_s_expression("(uvw)"),
            id(s_exp.children[1]): ns.parse_s_expression("(abc (def (ghi)))"),
        }
        replaced = s_exp.replace_nodes_by_id(id_to_node)
        self.assertEqual(
            ns.render_s_expression(replaced), "(foo (uvw) (abc (def (ghi))))"
        )

    def test_with_duplicates(self):
        s_exp = ns.parse_s_expression("(foo (bar) (bar) (bar))")
        id_to_node = {
            id(s_exp.children[0]): ns.parse_s_expression("(uvw)"),
            id(s_exp): ns.parse_s_expression("(abc (def (ghi)))"),
        }
        replaced = s_exp.replace_nodes_by_id(id_to_node)
        self.assertEqual(ns.render_s_expression(replaced), "(abc (def (ghi)))")



"""
Type signatures are used to represent the types of functions in the
DSL.
"""

import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from itertools import product
from typing import Callable, List, Tuple

import numpy as np
from frozendict import frozendict

from neurosym.types.type import (
    ArrowType,
    ListType,
    Type,
    TypeVariable,
    UnificationError,
)
from neurosym.types.type_with_environment import Environment, TypeWithEnvironment


class TypeSignature(ABC):
    """
    Represents a type signature, which is a function converting back and
        forth between types (outputs) and lists of types (inputs).
    """

    @abstractmethod
    def unify_return(self, twe: TypeWithEnvironment) -> List[TypeWithEnvironment]:
        """
        Returns a list of types, one for each of the arguments, or None
        if the type cannot be unified.
        """

    @abstractmethod
    def return_type_template(self) -> Type:
        """
        Returns a template for the return type, with type variables.

        This can be an over-approximation of the actual return type,
            but it must not be an under-approximation.
        """

    @abstractmethod
    def unify_arguments(self, twes: List[TypeWithEnvironment]) -> TypeWithEnvironment:
        """
        Returns the return type of the function, or None if the types
        cannot be unified.
        """

    @abstractmethod
    def arity(self) -> int:
        """
        Returns the arity of the function, i.e., the number of arguments.
        """

    @abstractmethod
    def render(self) -> str:
        """
        Render this type signature as a string.
        """


@dataclass
class FunctionTypeSignature(TypeSignature):
    """
    Represents a concrete type signature, where the return type is known and the
    arguments are known.
    """

    arguments: List[Type]
    return_type: Type

    @classmethod
    def from_type(cls, typ: Type) -> "FunctionTypeSignature":
        """
        Create a function type signature from a type, which must be an arrow type.
        """
        assert isinstance(typ, ArrowType)
        return cls(list(typ.input_type), typ.output_type)

    def unify_return(self, twe: TypeWithEnvironment) -> List[TypeWithEnvironment]:
        try:
            mapping = twe.typ.unify(self.return_type)
        except UnificationError:
            return None
        return [
            TypeWithEnvironment(t.subst_type_vars(mapping), twe.env)
            for t in self.arguments
        ]

    def return_type_template(self) -> Type:
        return self.return_type

    def unify_arguments(self, twes: List[TypeWithEnvironment]) -> TypeWithEnvironment:
        types = [x.typ for x in twes]
        envs = [x.env for x in twes]
        env = Environment.merge_all(*envs)
        mapping = {}
        try:
            for t1, t2 in zip(self.arguments, types):
                for_arg = t1.unify(t2)
                for k, v in for_arg.items():
                    if k in mapping:
                        if mapping[k] != v:
                            return None
                    else:
                        mapping[k] = v
        except UnificationError:
            return None
        return TypeWithEnvironment(self.return_type.subst_type_vars(mapping), env)

    def arity(self) -> int:
        return len(self.arguments)

    def astype(self) -> Type:
        """
        Convert this type signature to an arrow type, which is the
            type of the function that this signature represents.
        """
        return ArrowType(tuple(self.arguments), self.return_type)

    def render(self) -> str:
        # pylint: disable=cyclic-import
        from neurosym.types.type_string_repr import render_type

        return render_type(self.astype())


@dataclass
class LambdaTypeSignature(TypeSignature):
    """
    Represents the type signature of the lambda production.
    """

    input_types: List[ArrowType]

    def arity(self) -> int:
        # just the body
        return 1

    def function_arity(self) -> int:
        """
        Get the arity of the function represented by the lambda,
            i.e., the number of arguments.
        """
        return len(self.input_types)

    def render(self) -> str:
        # pylint: disable=cyclic-import
        from neurosym.types.type_string_repr import render_type

        body = TypeVariable("body")
        input_types = ";".join(render_type(x) for x in self.input_types)
        lambda_type = f"L<{render_type(body)}|{input_types}>"

        return f"{lambda_type} -> {render_type(ArrowType(self.input_types, body))}"

    def unify_return(self, twe: TypeWithEnvironment) -> List[TypeWithEnvironment]:
        if not isinstance(twe.typ, ArrowType):
            return None
        if twe.typ.input_type != self.input_types:
            return None
        return [
            TypeWithEnvironment(
                twe.typ.output_type,
                twe.env.child(*self.input_types),
            )
        ]

    def return_type_template(self) -> Type:
        return ArrowType(self.input_types, TypeVariable("body"))

    def unify_arguments(self, twes: List[TypeWithEnvironment]) -> TypeWithEnvironment:
        if len(twes) != 1:
            return None
        parent = twes[0].env.parent(self.input_types)
        return TypeWithEnvironment(ArrowType(self.input_types, twes[0].typ), parent)


@dataclass
class VariableTypeSignature(TypeSignature):
    """
    Represents the type signature of variable production.
    """

    variable_type: Type
    index_in_env: int

    def arity(self) -> int:
        # leaf
        return 0

    def render(self) -> str:
        # pylint: disable=cyclic-import
        from neurosym.types.type_string_repr import render_type

        return f"V<{render_type(self.variable_type)}@{self.index_in_env}>"

    def unify_return(self, twe: TypeWithEnvironment) -> List[TypeWithEnvironment]:
        if twe.typ != self.variable_type:
            return None
        if not twe.env.contains_type_at(self.variable_type, self.index_in_env):
            return None
        return []

    def return_type_template(self) -> Type:
        return self.variable_type

    def unify_arguments(self, twes: List[TypeWithEnvironment]) -> TypeWithEnvironment:
        if len(twes) != 0:
            return None
        return TypeWithEnvironment(
            self.variable_type,
            Environment(frozendict({self.index_in_env: self.variable_type})),
        )


def bottom_up_enumerate_types(
    terminals: List[Type],
    constructors: List[Tuple[int, Callable]],
    max_expansion_steps=np.inf,
    max_overall_depth=np.inf,
):
    """
    Returns a list of all possible expansions of the given terminals
    using the given constructors.
    """
    assert (
        min(max_expansion_steps, max_overall_depth) < np.inf
    ), "must specify either max_expansion_steps or max_overall_depth"
    current_with_depth = [(t, t.depth, 0) for t in terminals]
    current_with_depth = [
        (t, d, n)
        for t, d, n in current_with_depth
        if d <= max_overall_depth and n <= max_expansion_steps
    ]
    overall = set()
    while True:
        overall.update(current_with_depth)
        current_with_depth = []
        for arity, fn in constructors:
            additional_depth = np.log2(arity + 1)
            will_work = [
                (t, d, n)
                for t, d, n in overall
                if d + additional_depth <= max_overall_depth
                and n + 1 <= max_expansion_steps
            ]
            for subentities in itertools.product(will_work, repeat=arity):
                types = [t for t, _, _ in subentities]
                depth = max((d for _, d, _ in subentities)) + additional_depth
                steps = max((n for _, _, n in subentities)) + 1
                assert depth <= max_overall_depth
                assert steps <= max_expansion_steps
                res = (fn(*types), depth, steps)
                if res in overall:
                    continue
                current_with_depth.append(res)
        if len(current_with_depth) == 0:
            break
    return sorted([t for t, _, _ in overall], key=str)


def signature_expansions(
    sig: FunctionTypeSignature,
    terminals: List[Type],
    constructors: List[Tuple[int, Callable]],
    max_expansion_steps=np.inf,
    max_overall_depth=np.inf,
):
    """
    Returns a list of all possible expansions of the given type signature.

    Any type variables that appear in both the arguments and return type
        will be kept, while any other type variables will be expanded.
    """
    variables_in_arguments = {
        var for arg in sig.arguments for var in arg.max_depth_per_type_variable()
    }
    variables_in_return = set(sig.return_type.max_depth_per_type_variable())
    exclude_variables = variables_in_arguments & variables_in_return
    return expansions(
        sig.astype(),
        terminals,
        constructors,
        max_expansion_steps,
        max_overall_depth,
        exclude_variables,
    )


def expansions(
    sig: Type,
    terminals: List[Type],
    constructors: List[Tuple[int, Callable]],
    max_expansion_steps=np.inf,
    max_overall_depth=np.inf,
    exclude_variables=(),
):
    """
    Returns a list of all possible expansions of the given type.
    """
    # pylint: disable=cyclic-import
    from neurosym.types.type_string_repr import render_type

    depth_by_var = sig.max_depth_per_type_variable()
    for var in exclude_variables:
        if var not in depth_by_var:
            raise ValueError(
                f"Variable {var} not in type signature {render_type(sig)}, "
                + f"cannot exclude. Valid options: {sorted(depth_by_var)}"
            )
        del depth_by_var[var]
    enumerations_by_var = {
        ty_var: bottom_up_enumerate_types(
            terminals,
            constructors,
            max_expansion_steps,
            max_overall_depth - depth_by_var[ty_var],
        )
        for ty_var in depth_by_var.keys()
    }
    variables = list(depth_by_var.keys())
    enumerations = [enumerations_by_var[ty_var] for ty_var in variables]
    for types in product(*enumerations):
        remap = dict(zip(variables, types))
        yield sig.subst_type_vars(remap)


def type_universe(types: List[Type], require_arity_up_to=None, no_zeroadic=False):
    """
    Produce a type universe from the given types.

    Returns a tuple of (atomic_types, constructors), where atomic_types
        represents the atomic types in the universe, and constructors
        is a list of tuples of the form (arity, constructor), where
        constructor is a function that takes `arity` types and
        produces a new type.
    """
    atomic_types = set()
    num_arrow_args = set()
    has_list = False
    for typ in types:
        for t in typ.walk_type_nodes():
            if t.is_atomic():
                atomic_types.add(t)
            if isinstance(t, ArrowType):
                num_arrow_args.add(len(t.input_type))
            if isinstance(t, ListType):
                has_list = True
    atomic_types = sorted(atomic_types, key=str)
    if require_arity_up_to is not None:
        num_arrow_args |= set(range(1, require_arity_up_to + 1))
    if no_zeroadic:
        num_arrow_args -= {0}
    num_arrow_args = sorted(num_arrow_args)

    constructors = []
    if has_list:
        constructors.append((1, ListType))
    for n in num_arrow_args:
        constructors.append(
            (
                n + 1,
                lambda *args: ArrowType(tuple(args[:-1]), args[-1]),
            )
        )
    return atomic_types, constructors



import unittest

import neurosym as ns

from ..utils import assertDSL


class ProduceDslTest(unittest.TestCase):
    def setUp(self):
        self.dfa = ns.python_dfa()

    def test_produce_dsl_basic(self):
        program = ns.python_to_python_ast("x = x + 2; y = y + x + 2")
        subset = ns.PythonDSLSubset.from_programs(self.dfa, program, root="M")
        dsl = ns.create_python_dsl(ns.python_dfa(), subset, "M")
        assertDSL(
            self,
            dsl.render(),
            """
            /choiceseq~seqS~2 :: (S, S) -> seqS
            /seq~seqS~2 :: (S, S) -> seqS
            /splice~S :: seqS -> S
            /subseq~seqS~2 :: (S, S) -> seqS
            Add~O :: () -> O
            Assert~S :: (E, E) -> S
            Assign~S :: ([L], E, TC) -> S
            AsyncFor~S :: (L, E, seqS, seqS, TC) -> S
            AugAssign~S :: (L, O, E) -> S
            Await~E :: E -> E
            BinOp~E :: (E, O, E) -> E
            Constant~E :: (Const, ConstKind) -> E
            Delete~S :: [L] -> S
            Expr~S :: E -> S
            For~S :: (L, E, seqS, seqS, TC) -> S
            IfExp~E :: (E, E, E) -> E
            If~S :: (E, seqS, seqS) -> S
            List~L :: ([L], Ctx) -> L
            Load~Ctx :: () -> Ctx
            Module~M :: (seqS, [TI]) -> M
            NamedExpr~E :: (L, E) -> E
            Name~E :: (Name, Ctx) -> E
            Name~L :: (Name, Ctx) -> L
            Raise~S :: (E, E) -> S
            Return~S :: E -> S
            Slice~Slice :: (E, E, E) -> Slice
            Starred~E :: (E, Ctx) -> E
            Starred~L :: (L, Ctx) -> L
            Store~Ctx :: () -> Ctx
            Subscript~E :: (E, SliceRoot, Ctx) -> E
            Subscript~L :: (E, SliceRoot, Ctx) -> L
            Tuple~L :: ([L], Ctx) -> L
            UnaryOp~E :: (O, E) -> E
            While~S :: (E, seqS, seqS) -> S
            YieldFrom~E :: E -> E
            Yield~E :: E -> E
            _slice_content~SliceRoot :: E -> SliceRoot
            _slice_slice~SliceRoot :: Slice -> SliceRoot
            _starred_content~L :: L -> L
            _starred_starred~L :: L -> L
            const-&x:0~Name :: () -> Name
            const-&y:0~Name :: () -> Name
            const-None~ConstKind :: () -> ConstKind
            const-None~TC :: () -> TC
            const-i2~Const :: () -> Const
            list~_L_~1 :: L -> [L]
            list~_TI_~0 :: () -> [TI]
            """,
        )

    def test_produce_dsl_bad_type_annotate_length(self):
        program = ns.python_to_python_ast("x: List[int, int] = 2")
        print(ns.render_s_expression(program.to_ns_s_exp()))
        subset = ns.PythonDSLSubset.from_programs(self.dfa, program, root="M")
        dsl = ns.create_python_dsl(ns.python_dfa(), subset, "M")
        ta_lines = {x.strip() for x in dsl.render().split("\n") if "list~TA" in x}
        self.assertEqual(ta_lines, {"list~TA~2 :: (TA, TA) -> TA"})

    def test_create_dsl_with_type_annot(self):
        # just check it doesn't crash
        program = ns.python_to_python_ast("x:int = 5; y: float=5")
        subset = ns.PythonDSLSubset.from_programs(self.dfa, program, root="M")
        ns.create_python_dsl(ns.python_dfa(), subset, "M")



from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Tuple

import torch
from torch import nn

from .base import BaseConfig


@dataclass
class MLPConfig(BaseConfig):
    input_size: int
    hidden_size: int
    output_size: int
    dropout: float = 0.0
    bias: bool = True
    nonlinearity: str = "LeakyReLU"
    loss: str = "MSELoss"


class MLP(nn.Sequential):
    """Simple MLP module."""

    def __init__(self, config: MLPConfig):
        self.config = config
        nonlinearity = getattr(nn, config.nonlinearity)
        model_dict = OrderedDict(
            fc1=nn.Linear(config.input_size, config.hidden_size, bias=config.bias),
            act1=nonlinearity(),
            dp1=nn.Dropout(p=config.dropout),
            fc3=nn.Linear(config.hidden_size, config.output_size, bias=config.bias),
        )
        super().__init__(model_dict)

        @torch.no_grad()
        def init_weights(m):
            if isinstance(m, nn.Linear):
                m.weight = nn.init.xavier_normal_(
                    m.weight, gain=torch.nn.init.calculate_gain("leaky_relu")
                )

        self.apply(init_weights)


def mlp_factory(**kwargs):
    """
    Allows instantiating an MLP module with a given input and output size.
    """

    def construct_model(input_shape: List[Tuple[int]], output_shape: Tuple[int]):
        assert len(input_shape) == 1, "MLP takes a single input only."
        input_size = input_shape[0][-1]
        output_size = output_shape[-1]
        cfg = MLPConfig(
            model_name="mlp",
            input_size=input_size,
            output_size=output_size,
            **kwargs,
        )
        return MLP(cfg)

    return construct_model



import ast
import json
import unittest
from functools import lru_cache

from increase_recursionlimit import increase_recursionlimit
from parameterized import parameterized

import neurosym as ns


class ParseUnparseInverseTest(unittest.TestCase):
    def canonicalize(self, python_code):
        return ast.unparse(ast.parse(python_code))

    def assert_valid_s_exp(self, s_exp, no_leaves):
        if not isinstance(s_exp, ns.SExpression):
            assert isinstance(s_exp, str)
            if no_leaves:
                self.fail(f"leaf: {s_exp}")
            return
        if s_exp.symbol not in {"/seq"}:
            self.assertTrue(isinstance(s_exp.symbol, str), repr(s_exp.symbol))
            if not no_leaves:
                self.assertTrue(len(s_exp.children) >= 1, repr(s_exp))
        for y in s_exp.children:
            self.assert_valid_s_exp(y, no_leaves)

    def check_s_exp(self, s_exp, no_leaves=False):
        self.maxDiff = None
        parsed = ns.s_exp_to_python_ast(s_exp)
        print(parsed)
        with increase_recursionlimit():
            s_exp_update = ns.render_s_expression(ns.parse_s_expression(s_exp))
        self.assertEqual(
            s_exp_update,
            ns.render_s_expression(parsed.to_ns_s_exp(dict(no_leaves=no_leaves))),
        )

    def check_with_args(self, test_code, no_leaves=False):
        test_code = self.canonicalize(test_code)
        s_exp = ns.python_to_s_exp(
            test_code, renderer_kwargs=dict(columns=80), no_leaves=no_leaves
        )
        with increase_recursionlimit():
            self.assert_valid_s_exp(ns.parse_s_expression(s_exp), no_leaves=no_leaves)
        self.check_s_exp(s_exp, no_leaves=no_leaves)
        s_exp_parsed = ns.s_exp_to_python_ast(s_exp, {})
        print(repr(s_exp_parsed))
        modified = ns.s_exp_to_python(s_exp)
        self.assertEqual(test_code, modified)

    def check(self, test_code):
        self.check_with_args(test_code)
        self.check_with_args(test_code, no_leaves=True)

    def test_basic_one_liners(self):
        self.check("x = 2")
        self.check("7")
        self.check("import abc")

    def test_imports(self):
        self.check("import os")
        self.check("from os import path")
        self.check("import os.path")

    def test_sequence_of_statements(self):
        self.maxDiff = None
        self.check("x = 2\ny = 3\nz = 4")
        self.assertEqual(
            ns.python_to_s_exp(
                "x = 2\ny = 3\nz = 4", renderer_kwargs=dict(columns=80000)
            ),
            ns.render_s_expression(
                ns.parse_s_expression(
                    """
                    (Module
                        (/seq
                            (Assign (list (Name &x:0 Store)) (Constant i2 None) None)
                            (Assign (list (Name &y:0 Store)) (Constant i3 None) None)
                            (Assign (list (Name &z:0 Store)) (Constant i4 None) None))
                        nil)
                    """
                )
            ),
        )

    def test_globals(self):
        self.assertEqual(
            ns.python_to_s_exp("import os", renderer_kwargs=dict(columns=80)),
            "(Module (/seq (Import (list (alias g_os None)))) nil)",
        )

    def test_builtins(self):
        self.check("print(True)")
        self.check("0")
        self.check("x = None")

    def test_if_expr(self):
        self.check("2 if x == 3 else 4")

    def test_strings(self):
        self.check("' '")
        self.check("x = 'abc '")
        self.check("x = 'a=é b=\\udc80 d=\U0010ffff'")
        self.check("x = '\\uABCD'")

    def test_lambda(self):
        self.check("lambda: 1 + 2")

    def test_subscript_basic(self):
        self.check("x[2]")

    def test_subscript_tuple(self):
        self.check("x[2, 3]")

    def test_if(self):
        self.check("if True: pass")

    def test_unparse_sequence(self):
        # should work with or without the Module wrapper
        self.assertEqual(
            ns.s_exp_to_python(
                "(Module (/seq (Assign (list (Name &x:0 Store)) (Constant i2 None) None)) nil)"
            ),
            "x = 2",
        )

        self.assertEqual(
            ns.s_exp_to_python(
                "(/seq (Assign (list (Name &x:0 Store)) (Constant i2 None) None))"
            ),
            "x = 2",
        )

    @parameterized.expand([(i,) for i in range(100)])
    def test_realistic(self, i):
        try:
            print(small_set_examples()[i])
            self.check(small_set_examples()[i])
        except Exception as e:
            self.assertFalse(f"Error: {e}")
            raise e

    def test_subscript_s_exp(self):
        self.check_s_exp(
            "(_slice_tuple (Tuple (list (_slice_content (Constant i3 None))) Load))"
        )


@lru_cache(None)
def small_set_examples():
    with open("test_data/small_set.json") as f:
        return json.load(f)



def fold_torch(func, l):  # noqa: E741
    """
    Arguments:
        func: (N, *C) -> (N, *C) -> (N, *C)
        l: List[(N, L, *C)]
    Returns:
        (N, *C)
    """
    result = l[:, 0, :]
    for i in range(1, l.shape[1]):
        result = func(result, l[:, i, :])
    return result


def map_torch(func, l):  # noqa: E741
    """
    Arguments:
        func: (N, *C1) -> (N, *C2)
        l: List[(N, L, *C1)]
    Returns:
        (N, L, *C2)
    """
    original_shape = l.shape
    reshaped = l.reshape(l.shape[0] * l.shape[1], *l.shape[2:])
    result = func(reshaped)
    return result.reshape(original_shape[0], original_shape[1], *result.shape[1:])
